
import irispie as ir
import irispie.series.functions as isf
from irispie.series._statistics import __all__ as __all__statistics
import numpy as np
import random as rn
import pytest

rn.seed(0)

span = ir.qq(2021,1) >> ir.qq(2025,4)

x1 = ir.Series(periods=span, func=rn.gauss, )
x10 = ir.Series(periods=span, func=rn.gauss, num_variants=10, )

extra_args_functions = ("percentile", "quantile", "nanpercentile", "nanquantile", )

parameter_names = ("func_name", )
parameter_values = [ (n, ) for n in __all__statistics if n not in extra_args_functions ]

@pytest.mark.parametrize(parameter_names, parameter_values, )
def test_statistics(func_name, ):
    # Run the method on the copy of the Series object
    y21 = x10.copy()
    getattr(y21, func_name, )()
    # Run the function creating a new Series object
    y22 = getattr(isf, func_name, )(x10, )
    # Run the NP function on the Series data
    expected_data = getattr(np, func_name, )(x10.data, axis=1, ).reshape(-1, 1, )
    #
    assert np.all(y21.data == expected_data)
    assert np.all(y22.data == expected_data)


def test_percentile_axis_1():
    q = (0, 50, 100, )
    # Run the method on the copy of the Series object
    y21 = x10.copy()
    y21.percentile(q, )
    # Run the function creating a new Series object
    y22 = isf.percentile(x10, q, )
    # Run the NP function on the Series data
    expected_data = np.percentile(x10.data, q, axis=1, ).T.reshape(-1, len(q), )
    #
    assert np.all(y21.data == expected_data)
    assert np.all(y22.data == expected_data)


def test_percentile_axis_0():
    q = (0, 50, 100, )
    # Run the function creating a new Series object
    y22 = isf.percentile(x10, q, axis=0, )
    # Run the NP function on the Series data
    expected_data = np.percentile(x10.data, q, axis=0, ).T.tolist()
    #
    assert all(y == e for y, e in zip(y22, expected_data))


def test_quantiles():
    q = (0, 0.5, 1, )
    # Run the method on the copy of the Series object
    y21 = x10.copy()
    y21.quantile(q, )
    # Run the function creating a new Series object
    y22 = isf.quantile(x10, q, )
    # Run the NP function on the Series data
    expected_data = np.quantile(x10.data, q, axis=1, ).T.reshape(-1, len(q), )
    #
    assert np.all(y21.data == expected_data)
    assert np.all(y22.data == expected_data)


def test_quantile_axis_0():
    q = (0, 0.5, 1, )
    # Run the function creating a new Series object
    y22 = isf.quantile(x10, q, axis=0, )
    # Run the NP function on the Series data
    expected_data = np.quantile(x10.data, q, axis=0, ).T.tolist()
    #
    assert all(y == e for y, e in zip(y22, expected_data))





"""
"""


#[
from __future__ import annotations

import copy as _cp
import functools as _ft
import itertools as _it
import numpy as _np
import re as _re

from ..conveniences import descriptions as _descriptions
from ..fords import descriptors as _descriptors
from ..equators import plain as _equators
from .. import equations as _equations
from ..equations import EquationKind, Equation
from .. import quantities as _quantities
from ..quantities import QuantityKind, Quantity
from .. import wrongdoings as _wrongdoings
from .. import makers as _makers
from .. import contexts as _contexts
from ..incidences.main import ZERO_SHIFT_TOKEN_PATTERN
from .. import sources as _sources
from .. sources import ModelSource

from ._flags import Flags

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Self, Callable, NoReturn, Any
    from numbers import Real

#]


_DEFAULT_STD_LINEAR = 1
_DEFAULT_STD_NONLINEAR = 0.01


_PLAIN_EQUATOR_EQUATION = (
    EquationKind.TRANSITION_EQUATION
    | EquationKind.MEASUREMENT_EQUATION
)


_DEFAULT_STD_NAME_FORMAT = "std_{}"
_DEFAULT_STD_DESCRIPTION_FORMAT = "(Std) {}"


_STD_PREFIX = "std_"
_STD_DESCRIPTION_PREFIX = "(Std) "


def is_std_name(name: str, /, ) -> bool:
    return name.startswith(_STD_PREFIX)


def std_name_from_shock_name(shock_name: str, /, ) -> str:
    return f"{_STD_PREFIX}{shock_name}"


def shock_name_from_std_name(std_name: str, /, ) -> str:
    if not is_std_name(std_name):
        raise ValueError(f"Invalid std name: {std_name}")
    return std_name.removeprefix(_STD_PREFIX, )


def std_description_from_shock_description(shock_description: str, /, ) -> str:
    return f"{_STD_DESCRIPTION_PREFIX}{shock_description}"


class Invariant(
    _descriptions.DescriptionMixin,
):
    """
    Invariant part of a Simultaenous object
    """
    #[

    _serialized_slots = (
        "quantities",
        "dynamic_equations",
        "steady_equations",
        "shock_qid_to_std_qid",
        "_context",
        "_default_std",
        "_flags",
        "_min_shift",
        "_max_shift",
        "__description__",
    )

    _derived_slots = (
        "dynamic_descriptor",
        "steady_descriptor",
        "update_steady_autovalues_in_variant",
        "_plain_dynamic_equator",
        "_plain_steady_equator",
    )

    __slots__ = _serialized_slots + _derived_slots

    def __init__(self, **kwargs, ) -> None:
        """
        """
        for n in self._serialized_slots:
            setattr(self, n, kwargs.get(n, ), )
        for n in self._derived_slots:
            setattr(self, n, None, )

    @classmethod
    def from_source(
        klass,
        source: ModelSource,
        /,
        check_syntax: bool = True,
        autodeclare_as: str | None = None,
        default_std: Real | None = None,
        description: str | None = None,
        **kwargs,
    ) -> Self:
        """
        """
        self = klass()
        self.set_description(
            description if description is not None
            else source.description
        )
        self._flags = Flags.from_kwargs(**kwargs, )
        self._default_std = _resolve_default_std(default_std, self._flags, )
        self._context = (dict(source.context) or {}) | {"__builtins__": None}
        #
        self.quantities = tuple(source.quantities)
        self.dynamic_equations = tuple(_equations.generate_equations_of_kind(
            source.dynamic_equations,
            kind=EquationKind.ENDOGENOUS_EQUATION | EquationKind.STEADY_AUTOVALUES,
        ))
        self.steady_equations = tuple(_equations.generate_equations_of_kind(
            source.steady_equations,
            kind=EquationKind.ENDOGENOUS_EQUATION | EquationKind.STEADY_AUTOVALUES,
        ))
        #
        # Create std_ parameters for all shocks
        if self._flags.is_stochastic:
            get_shocks = _ft.partial(_quantities.generate_quantities_of_kind, self.quantities, )
            unanticipated_shocks = get_shocks(kind=_quantities.UNANTICIPATED_SHOCK, )
            anticipated_shocks = get_shocks(kind=_quantities.ANTICIPATED_SHOCK, )
            measurement_shocks = get_shocks(kind=_quantities.MEASUREMENT_SHOCK, )
            #
            entry = len(self.quantities)
            self.quantities += tuple(
                _create_std_for_shock(i, QuantityKind.UNANTICIPATED_STD, entry, )
                for i in unanticipated_shocks
            )
            self.quantities += tuple(
                _create_std_for_shock(i, QuantityKind.ANTICIPATED_STD, entry, )
                for i in anticipated_shocks
            )
            self.quantities += tuple(
                _create_std_for_shock(i, QuantityKind.MEASUREMENT_STD, entry, )
                for i in measurement_shocks
            )
        #
        # Look up undeclared names; autodeclare these names or throw an error
        undeclared_names = _collect_undeclared_names(
            self.quantities,
            self.dynamic_equations + self.steady_equations,
        )
        if undeclared_names and autodeclare_as is None:
            raise _wrongdoings.IrisPieCritical([
                "These names are used in equations but not declared:",
                *undeclared_names,
                ])
        self.quantities += tuple(_generate_quantities_for_undeclared_names(
            undeclared_names, autodeclare_as, entry=len(self.quantities),
        ))
        #
        # Verify uniqueness of all quantity names
        _quantities.check_unique_names(self.quantities)
        #
        # Count of transition equations must equal count of transition variables
        _check_numbers_of_variables_equations(
            self.quantities,
            self.dynamic_equations,
            QuantityKind.TRANSITION_VARIABLE,
            EquationKind.TRANSITION_EQUATION,
        )
        #
        # Number of measurement equations must equal number of measurement variables
        _check_numbers_of_variables_equations(
            self.quantities,
            self.dynamic_equations,
            QuantityKind.MEASUREMENT_VARIABLE,
            EquationKind.MEASUREMENT_EQUATION,
        )
        #
        self.quantities = _quantities.reorder_by_kind(self.quantities, )
        self.dynamic_equations, self.steady_equations = _reorder_equations_by_kind(self.dynamic_equations, self.steady_equations, )
        _quantities.stamp_id(self.quantities, )
        _equations.stamp_id(self.dynamic_equations, )
        _equations.stamp_id(self.steady_equations, )
        #
        # For stochastic models, create a mapping from shock qid to std qid
        self.shock_qid_to_std_qid = (
            _create_shock_qid_to_std_qid(self.quantities, )
            if not self._flags.is_deterministic else {}
        )
        #
        # Finalize equations by replacing names with qid pointers
        name_to_qid = _quantities.create_name_to_qid(self.quantities, )
        _equations.finalize_equations(self.dynamic_equations, name_to_qid, )
        _equations.finalize_equations(self.steady_equations, name_to_qid, )
        #
        if check_syntax:
            _check_syntax(self.dynamic_equations, self._context, )
            _check_syntax(self.steady_equations, self._context, )
        #
        self._populate_min_max_shifts()
        #
        self._populate_derived_attributes()
        #
        return self

    def _populate_derived_attributes(self, /, ) -> None:
        """
        """
        #
        # Descriptors for first-order systems
        self.dynamic_descriptor = _descriptors.Descriptor(self.dynamic_equations, self.quantities, self._context, )
        self.steady_descriptor = _descriptors.Descriptor(self.steady_equations, self.quantities, self._context, )
        #
        # Evaluators of LHS-minus-RHS in dynamic and steady equations
        self._plain_dynamic_equator = _equators.PlainEquator(
            _equations.generate_equations_of_kind(self.dynamic_equations, _PLAIN_EQUATOR_EQUATION, ),
            context=self._context,
        )
        self._plain_steady_equator = _equators.PlainEquator(
            _equations.generate_equations_of_kind(self.steady_equations, _PLAIN_EQUATOR_EQUATION, ),
            context=self._context,
        )
        #
        # Steady autovalue updater
        steady_autovalues = tuple(_equations.generate_equations_of_kind(
            self.steady_equations,
            kind=EquationKind.STEADY_AUTOVALUES,
        ))
        _create_steady_autovalue_updater(self, steady_autovalues, )

    def copy(self, /, ) -> Self:
        """
        """
        # TODO: Optimize
        return _cp.deepcopy(self, )

    @property
    def num_transition_equations(self, /, ) -> int:
        """
        """
        return sum(
            1 for e in self.dynamic_equations
            if e.kind is EquationKind.TRANSITION_EQUATION
        )

    @property
    def num_measurement_equations(self, /, ) -> int:
        """
        """
        return sum(
            1 for e in self.dynamic_equations
            if e.kind is EquationKind.MEASUREMENT_EQUATION
        )

    def _populate_min_max_shifts(self, /, ) -> None:
        """
        """
        self._min_shift = _equations.get_min_shift_from_equations(
            self.dynamic_equations + self.steady_equations,
        )
        self._max_shift = _equations.get_max_shift_from_equations(
            self.dynamic_equations + self.steady_equations,
        )

    def __getstate__(self, /, ) -> dict[str, Any]:
        return {
            k: getattr(self, k)
            for k in self._serialized_slots
        }

    def __setstate__(self, state: dict[str, Any], ) -> None:
        for k in self._serialized_slots:
            setattr(self, k, state[k])
        self._populate_derived_attributes()

    def to_portable(self, /, ) -> dict[str, Any]:
        return {
            "description": str(self.get_description()),
            "flags": self._flags.to_portable(),
            "quantities": _quantities.to_portable(self.quantities, ),
            "equations": _equations.to_portable(self.dynamic_equations, self.steady_equations, ),
            "context": _contexts.to_portable(self._context, ),
        }

    @classmethod
    def from_portable(klass, portable: dict[str, Any], /, ) -> Self:
        description = str(portable["description"], )
        flags = portable["flags"]
        quantities = _quantities.from_portable(portable["quantities"], )
        dynamic_equations, steady_equations = _equations.from_portable(portable["equations"], )
        context = _contexts.from_portable(portable["context"], )
        source = ModelSource(
            quantities=quantities,
            dynamic_equations=dynamic_equations,
            steady_equations=steady_equations,
            context=_contexts.from_portable(portable["context"], ),
            description=description,
            **flags,
        )
        return klass.from_source(source, check_syntax=False, )

    #]


def _check_syntax(equations, function_context, /, ):
    """
    Try all equations at once; if this fails, do equation by equation to catch the troublemakers
    """
    try:
        eval(_equations.create_equator_func_string(equations, ), )
    except:
        _catch_troublemakers(equations, function_context, )
    #]


def _catch_troublemakers(equations, function_context, /, ):
    """
    Catch the troublemakers
    """
    #[
    fail = [
        eqn.human for eqn in equations
        if not _success_creating_lambda(eqn, function_context)
    ]
    if fail:
        message = ["Syntax error in these equations"] + fail
        raise _wrongdoings.IrisPieCritical(message, )
    #]


def _success_creating_lambda(equation, function_context):
    """
    """
    #[
    try:
        eval(_equations.create_equator_func_string([equation.xtring]))
        return True
    except Exception as ex:
        return False
    #]


def _create_shock_qid_to_std_qid(
    quantities: Iterable[Quantity],
    /,
) -> dict[int, int]:
    """
    """
    #[
    name_to_qid = _quantities.create_name_to_qid(quantities, )
    qid_to_name = _quantities.create_qid_to_name(quantities, )
    kind = QuantityKind.ANY_SHOCK
    all_shock_qids = tuple(_quantities.generate_qids_by_kind(quantities, kind))
    return {
        shock_qid: name_to_qid[std_name_from_shock_name(qid_to_name[shock_qid])]
        for shock_qid in all_shock_qids
    }
    #]


def _create_std_for_shock(
    shock: Quantity,
    std_kind: QuantityKind,
    entry: int,
) -> Quantity:
    """
    """
    #[
    std_name = std_name_from_shock_name(shock.human, )
    std_description = std_description_from_shock_description(shock.description or shock.human, )
    return Quantity(
        id=None,
        human=std_name,
        kind=std_kind,
        logly=None,
        description=std_description,
        entry=entry,
    )
    #]


def _collect_undeclared_names(
    quantities: tuple[Quantity, ...],
    equations: tuple[Equation, ...],
    /,
) -> tuple[str, ...]:
    """
    """
    #[
    all_names = set(_quantities.generate_all_quantity_names(quantities, ))
    all_names_in_equations = set(_equations.generate_all_names_from_equations(equations, ))
    return tuple(all_names_in_equations - all_names)
    #]


def _generate_quantities_for_undeclared_names(
    undeclared_names: set[str],
    autodeclare_as: str | None,
    /,
    *,
    entry: int,
) -> tuple[Quantity, ...] | NoReturn:
    """
    """
    #[
    if not undeclared_names:
        return tuple()
    autodeclare_kind = QuantityKind.from_keyword(autodeclare_as, )
    return (
        Quantity(
            id=None,
            human=name,
            kind=autodeclare_kind,
            logly=None,
            description=None,
            entry=entry,
        )
        for name in undeclared_names
    )
    #]


def _check_numbers_of_variables_equations(
    quantities: tuple[Quantity, ...],
    equations: tuple[Equation, ...],
    quantity_kind: QuantityKind,
    equation_kind: EquationKind,
    /,
) -> None:
    """
    """
    #[
    num_variables = _quantities.count_quantities_of_kind(quantities, quantity_kind, )
    num_equations = _equations.count_equations_of_kind(equations, equation_kind, )
    if num_variables != num_equations:
        raise _wrongdoings.IrisPieCritical([
            "Inconsistent numbers of variables and equations",
            f"Number of {quantity_kind.human.lower()}s: {num_variables}",
            f"Number of {equation_kind.human.lower()}s: {num_equations}",
        ])
    #]


def _resolve_default_std(
    custom_default_std: Real | None,
    flags: Flags,
) -> Real:
    """
    """
    #[
    if custom_default_std is not None:
        return float(custom_default_std)
    elif flags.is_linear:
        return _DEFAULT_STD_LINEAR
    else:
        return _DEFAULT_STD_NONLINEAR
    #]


_STEADY_AUTOVALUE_PATTERN = _re.compile(
    r"-\(" + ZERO_SHIFT_TOKEN_PATTERN + r"\)(.*)"
)


def _create_steady_autovalue_updater(
    self,
    steady_autovalues: Iterable[Equation],
    /,
) -> tuple[int, str]:
    """
    """
    #[
    lhs_qids = []
    rhs_xtrings = []
    for i in steady_autovalues:
        match = _STEADY_AUTOVALUE_PATTERN.match(i.xtring, )
        lhs_qid = int(match.group(1))
        rhs_xtring = match.group(2)
        lhs_qids.append(lhs_qid, )
        rhs_xtrings.append(rhs_xtring, )

    if not lhs_qids:
        self.update_steady_autovalues_in_variant = None
        return

    joined_rhs_xtrings = "(" + "  ,  ".join(rhs_xtrings, ) + " , )"
    func, func_str, *_ = _makers.make_function(
        "__equator",
        _equators.EQUATOR_ARGS,
        joined_rhs_xtrings,
        self._context,
    )
    num_columns = self._max_shift - self._min_shift + 1 
    shift_in_first_column = self._min_shift
    t = -self._min_shift
    qid_to_logly = _quantities.create_qid_to_logly(self.quantities, )

    def update_steady_autovalues(variant, ):
        steady_array = variant.create_steady_array(
            qid_to_logly,
            num_columns=num_columns,
            shift_in_first_column=shift_in_first_column,
        )
        values = _np.vstack(func(steady_array, t, ), )
        variant.update_levels_from_array(values, lhs_qids, )

    self.update_steady_autovalues_in_variant = update_steady_autovalues

    #]


def _reorder_equations_by_kind(
    dynamic_equations: tuple[Equation],
    steady_equations: tuple[Equation],
) -> tuple[tuple[Equation], tuple[Equation]]:
    """
    Reorder dynamic and steady equations at the same time to guarantee the same order for both
    """
    #[
    zipped = zip(dynamic_equations, steady_equations)
    zipped = sorted(zipped, key=lambda x: x[0].kind.value, )
    return zip(*zipped)
    #]





import pytest
import random as rn
import numpy as np
import irispie as ir


rn.seed(0)


def sum_round(x, tolerance=10, ):
    return np.round(x, tolerance).sum()


def random_series():
    return ir.Series(periods=ir.qq(2020,1)>>ir.qq(2025,4), func=rn.gauss, )


N = 10


@pytest.mark.parametrize(
    ["x"], [ (random_series(), ) for _ in range(N, ) ]
)
def test_hpf_additive(x, ):
    xt, xg = ir.hpf(x, )
    assert sum_round((xt + xg).get_data() - x.get_data()) == 0


@pytest.mark.parametrize(
    ["x", "level_value"], [(random_series(), rn.uniform(0, 1)) for _ in range(N, )]
)
def test_hpf_level(x, level_value, ):
    x = ir.Series(periods=ir.qq(2020,1)>>ir.qq(2025,4), func=rn.gauss, )
    level_date = x.end
    level = ir.Series(periods=level_date, values=level_value, )
    xt, xg = ir.hpf(x, level=level, )
    assert sum_round(xt[level_date] - level_value, 0) == 0


@pytest.mark.parametrize(
    ["x", "change_value"], [(random_series(), rn.uniform(0, 1)) for _ in range(N, )]
)
def test_hpf_change(x, change_value, ):
    x = ir.Series(periods=ir.qq(2020,1)>>ir.qq(2025,4), func=rn.gauss, )
    change_date = x.end
    change = ir.Series(periods=change_date, values=change_value, )
    xt, xg = ir.hpf(x, change=change, )
    diff_xt = ir.diff(xt)
    assert sum_round(diff_xt[change_date] - change_value, 0) == 0





"""
"""


#[
from __future__ import annotations

from typing import (Callable, )
import numpy as _np
import functools as _ft

from ..frames import (Frame, )
from ..dataslates.main import (Dataslate, )
from .solutions import (Solution, )
#]


def _simulate_anticipated_shocks(
    model_v,
    dataslate_v: Dataslate,
    frame: Frame,
    get_solution_expansion: Callable,
) -> list[_np.ndarray | None]:
    """
    """
    #[
    if model_v.num_variants != 1:
        raise ValueError("Simulator requires a singleton simulatable object")
    #
    solution = model_v.get_solution()
    solution_vectors = model_v.solution_vectors
    base_columns = dataslate_v.base_columns
    first_column = frame.start - dataslate_v.periods[0]
    last_column = frame.simulation_end - dataslate_v.periods[0]
    columns_to_run = tuple(range(first_column, last_column+1))
    working_data = dataslate_v.get_data_variant(0, )
    anticipated_shocks = extract_shock_values(working_data, solution_vectors.anticipated_shocks, )
    #
    impact = [None] * dataslate_v.num_periods
    column_incidence = _np.any(anticipated_shocks[:, columns_to_run], axis=0, ).tolist()
    if any(column_incidence, ):
        #
        # Last shock is at t+forward
        forward = len(column_incidence) - column_incidence[::-1].index(True, ) - 1
        Rx = get_solution_expansion(solution, forward, )
        #
        # Last column within dataslate_v
        last_ant_column = (
            first_column + forward
            if forward is not None
            else first_column - 1
        )
        #
        for t in columns_to_run:
            impact[t] = sum(
                Rx[k] @ anticipated_shocks[:, s]
                for k, s in enumerate(range(t, last_ant_column+1))
            )
    return impact
    #]


simulate_square_anticipated_shocks = _ft.partial(
    _simulate_anticipated_shocks,
    get_solution_expansion=Solution.expand_square_solution,
)


simulate_triangular_anticipated_shocks = _ft.partial(
    _simulate_anticipated_shocks,
    get_solution_expansion=Solution.expand_triangular_solution,
)


def extract_shock_values(working_data, shock_vector, ) -> _np.ndarray:
    """
    """
    #[
    shock_qids = [t.qid for t in shock_vector]
    return working_data[shock_qids, :]
    #]




"""
"""


#[

from __future__ import annotations

from .. import dates as _dates
from ..dates import Period
from ..fords.descriptors import SolutionVectors
from ..simultaneous.main import Simultaneous
from .. import quantities as _quantities
from ..quantities import QuantityKind, Quantity
from ..incidences.main import Token

#]


class Invariant:
    """
    """
    #[

    __slots__ = (
        "num_periods",
        "quantities",
        "max_lag",
        "max_lead",
        "source_solution_vectors",
        "stacked_solution_vectors",
        "index_xi",
        "index_u",
        "index_v",
        "index_y",
        "index_w",
    )

    def __init__(self, /, **kwargs) -> None:
        """
        """
        for slot in self.__slots__:
            setattr(self, slot, kwargs.get(slot, None, ), )

    @classmethod
    def from_simultaneous(
        klass,
        model: Simultaneous,
        num_periods: int,
        *,
        transition_variables: None | Iterable[str] = None,
        anticipated_shocks: None | Iterable[str] = None,
        unanticipated_shocks: None | Iterable[str] = None,
        measurement_variables: None | Iterable[str] = None,
        measurement_shocks: None | Iterable[str] = None,
    ) -> Self:
        """
        """
        self = klass()
        self.num_periods = num_periods
        self.quantities = model.quantities
        self.max_lag = model.max_lag
        self.max_lead = model.max_lead
        self.source_solution_vectors = model._invariant.dynamic_descriptor.solution_vectors
        #
        name_to_qid = model.create_name_to_qid()
        quantity_names = {
            "transition_variables": transition_variables,
            "anticipated_shocks": anticipated_shocks,
            "unanticipated_shocks": unanticipated_shocks,
            "measurement_variables": measurement_variables,
            "measurement_shocks": measurement_shocks,
        }
        for kind in quantity_names.keys():
            if quantity_names[kind] is None:
                quantity_names[kind] = model.get_names(kind=QuantityKind.from_keyword(kind, ), )
        #
        quantity_ids = { kind: None for kind in quantity_names.keys() }
        self.stacked_solution_vectors = SolutionVectors()
        for kind in quantity_names.keys():
            sorted_qids = sorted(set(tuple(name_to_qid[name] for name in quantity_names[kind])))
            vector = tuple( Token(qid, 0, ) for qid in sorted_qids )
            setattr(self.stacked_solution_vectors, kind, vector, )
        #
        self._populate_indexes()
        #
        return self

    def _populate_indexes(self, /, ) -> None:
        self.index_xi = _create_index(
            self.source_solution_vectors.transition_variables,
            self.stacked_solution_vectors.transition_variables,
        )
        self.index_u = _create_index(
            self.source_solution_vectors.unanticipated_shocks,
            self.stacked_solution_vectors.unanticipated_shocks,
        )
        self.index_v = _create_index(
            self.source_solution_vectors.anticipated_shocks,
            self.stacked_solution_vectors.anticipated_shocks,
        )
        self.index_y = _create_index(
            self.source_solution_vectors.measurement_variables,
            self.stacked_solution_vectors.measurement_variables,
        )
        self.index_w = _create_index(
            self.source_solution_vectors.measurement_shocks,
            self.stacked_solution_vectors.measurement_shocks,
        )

    def get_base_periods(self, start, / ) -> tuple[Period]:
        end = start + (self.num_periods - 1)
        return _dates.periods_from_until(start, end, )

    #]


def _create_index(source_vector, stacked_vector, /, ) -> list[int]:
    return [ source_vector.index(tok) for tok in stacked_vector ]





"""
"""


#[

from __future__ import annotations
from typing import Any

#]


_ContextType = dict[str, Any]
_PortableType = dict[str, None]


def to_portable(context: _ContextType) -> _PortableType:
    """
    """
    return {
        k: None
        for k in context.keys()
        if k != "__builtins__"
    }

def from_portable(portable: _PortableType) -> _ContextType:
    """
    """
    keys = tuple(portable.keys()) + ("__builtins__", )
    return { k: None for k in keys }





import pytest
import irispie as ir
import random as rn


rn.seed(0)

_ROUND_TO = 5

_SOLVER_SETTINGS = {
    "step_tolerance": 100,
}

_START_SIM = ir.qq(2020,1)
_END_SIM =  ir.qq(2020,4)

_INPUT_DB = ir.Databox()
_INPUT_DB["y"] = ir.Series(
    periods=_START_SIM-1>>_END_SIM,
    func=lambda: 100 + rn.uniform(0, 1)/10,
)
_INPUT_DB["x"] = ir.Series(
    periods=_START_SIM-1>>_END_SIM,
    values=100,
)
_INPUT_DB["x"][_START_SIM-1] = _INPUT_DB["x"][_START_SIM-1]



@pytest.mark.parametrize(
    ["func"], [ ("pct", ), ("log", ), ("diff", ), ("roc", ), ("diff_log", ) ],
)
def test_simulate(func: str ):

    source = r"""
    !transition_variables
        x
    !exogenous_variables
        y
    !equations
        {{func}}(x) = {{func}}(y);
    """

    m = ir.Simultaneous.from_string(
        source,
        context={"func": func},
    )

    s = m.simulate(
        _INPUT_DB,
        _START_SIM>>_END_SIM,
        method="period_by_period",
        solver_settings=_SOLVER_SETTINGS,
    )

    assert ir.round(s["x"], _ROUND_TO) == ir.round(s["y"], _ROUND_TO)





"""
# Stacked-time linear system

$$
\tilda x_t = 
$$

"""


#[

from __future__ import annotations

import numpy as _np

from ..fords.solutions import Solution

from ._invariants import Invariant

#]


mat_pow = _np.linalg.matrix_power


class Variant:
    """
    """

    __slots__ = (
        "TT", "PP", "RR", "KK",
        "AA", "BB", "CC", "DD", "HH",
    )

    def __init__(self, ) -> None:
        """
        """
        for n in self.__slots__:
            setattr(self, n, None, )

    @classmethod
    def from_solution(
        klass,
        invariant: Invariant,
        solution: Solution,
    ) -> Self:
        """
        """
        self = klass()
        num_periods = invariant.num_periods
        T, P, _, K, Z, H, D, = solution.unpack_square_solution()
        forward = num_periods - 1
        R = solution.expand_square_solution(forward=forward, )
        index_xi = invariant.index_xi
        index_u = invariant.index_u
        index_v = invariant.index_v
        index_y = invariant.index_y
        index_w = invariant.index_w
        num_xi = T.shape[0]
        num_u = P.shape[1]
        num_v = R[0].shape[1]
        num_y_included = len(index_y)
        num_u_included = len(index_u)
        num_v_included = len(index_v)
        num_w_included = len(index_w)
        Px = _np.hstack([P[:, index_u], _np.zeros((num_xi, num_u_included*(num_periods-1)))], )
        Rx = _np.hstack([r[:, index_v] for r in R])
        Hx = _np.hstack([H[index_y, :][:, index_w], _np.zeros((num_y_included, num_w_included*(num_periods-1)))], )
        Z = Z[index_y, :]
        H = H[index_y, :][: , index_w]
        D = D[index_y]
        #
        TT = []
        KK = []
        PP = []
        RR = []
        #
        AA = []
        BB = []
        CC = []
        DD = []
        HH = []
        #
        full_TT = _np.eye(num_xi, )
        full_KK = _np.zeros((num_xi, ), )
        full_PP = _np.zeros((num_xi, num_u_included*num_periods, ), )
        full_RR = _np.zeros((num_xi, num_v_included*num_periods, ), )
        for k in range(num_periods):
            full_TT = T @ full_TT
            full_KK = T @ full_KK + K
            full_PP = T @ full_PP + Px
            full_RR = T @ full_RR + Rx
            TT.append(full_TT[invariant.index_xi, :])
            KK.append(full_KK[invariant.index_xi])
            PP.append(full_PP[invariant.index_xi, :])
            RR.append(full_RR[invariant.index_xi, :])
            #
            AA.append(Z @ full_TT)
            BB.append(Z @ full_PP)
            CC.append(Z @ full_RR)
            DD.append(Z @ full_KK + D)
            HH.append(H)
            #
            Px = _move_forward(Px, num_u_included)
            Rx = _move_forward(Rx, num_v_included)
            Hx = _move_forward(Hx, num_w_included)
        #
        self.TT = _np.vstack(TT, )
        self.PP = _np.vstack(PP, )
        self.RR = _np.vstack(RR, )
        self.KK = _np.hstack(KK, )
        #
        self.AA = _np.vstack(AA, )
        self.BB = _np.vstack(BB, )
        self.CC = _np.vstack(CC, )
        self.DD = _np.hstack(DD, )
        self.HH = _np.vstack(HH, )
        #
        return self


def _move_forward(X: _np.ndarray, by: int, /, ) -> _np.ndarray:
    num_rows = X.shape[0]
    return _np.hstack([_np.zeros((num_rows, by)), X[:, :-by], ], )




"""
Main time series class definition
"""


#[

from __future__ import annotations

from numbers import Real
from collections.abc import Iterable, Callable
from typing import Self, Any, TypeAlias, NoReturn, Literal
from types import EllipsisType
import numpy as _np
import operator as _op
import documark as _dm

from ..conveniences import descriptions as _descriptions
from ..conveniences import copies as _copies
from ..conveniences import iterators as _iterators
from ..dates import (Period, Span, Frequency, EmptyRanger, )
from .. import dates as _dates
from .. import wrongdoings as _wrongdoings
from .. import has_variants as _has_variants

from ._functionalize import FUNC_STRING

from . import _indexing
from . import _plotly
from . import _views
from . import _temporal
from . import _filling
from . import _conversions
from . import _hp
from . import _extrapolate
from . import _x13
from . import _moving
from . import _elementwise
from . import _statistics

from ._categories import CATEGORIES

#]


#
# The following functions can return either a new time series object if applied
# across variants (axis=1), or a scalar or a list of values if applied across
# time (axis=0)
#


__all__ = ["Series", "shift", ]


Dates = Period | Iterable[Period] | Span | EllipsisType | None
VariantsRequestType = int | Iterable[int] | slice | None
LayMethodType = Literal["by_span", ]
ShiftType = int | Literal["yoy", "soy", "eopy", "tty", ]
AxisType = Literal[0, 1]


def _get_date_positions(dates, base, num_periods, /, ):
    pos = tuple(_dates.date_index(dates, base))
    min_pos = min((x for x in pos if x is not None), default=0)
    max_pos = max((x for x in pos if x is not None), default=0)
    add_before = max(-min_pos, 0)
    add_after = max(max_pos - num_periods + 1, 0)
    pos_adjusted = [
        p + add_before if p is not None else None
        for p in pos
    ]
    return pos_adjusted, add_before, add_after


@_dm.reference(
    path=("data_management", "time_series.md", ),
    categories=CATEGORIES,
)
class Series(
    _indexing.Inlay,
    _conversions.Inlay,
    _temporal.Inlay,
    _filling.Inlay,
    _hp.Inlay,
    _extrapolate.Inlay,
    _moving.Inlay,
    _x13.Inlay,
    _plotly.Inlay,
    _views.Inlay,
    _elementwise.Inlay,
    _statistics.Inlay,
    #
    _descriptions.DescriptionMixin,
    _copies.Mixin,
):
    r"""
················································································

Time series
============

Time `Series` objects represent numerical time series, organized as rows of
observations stored in [`numpy`](https://numpy.org) arrays and time stamped
using [time `Periods`](periods.md). A `Series` object can hold multiple
variants of the data, stored as mutliple columns.

················································································
    """
    #[

    __slots__ = (
        "start",
        "data",
        "data_type",
        "metadata",
        "__description__",
    )

    _numeric_format: str = "15g"
    _short_str_format: str = ">15"
    _date_str_format: str = ">12"
    _missing = _np.nan
    _missing_str: str = "⋅"
    _test_missing_period = staticmethod(lambda x: _np.all(_np.isnan(x)))

    @_dm.reference(
        category="constructor",
        call_name="Series",
    )
    def __init__(
        self,
        /,
        *,
        num_variants: int = 1,
        data_type: type = _np.float64,
        description: str = "",
        start: Period | None = None,
        start_date: Period | None = None,
        periods: Iterable[Period] | None = None,
        dates: Iterable[Period] | None = None,
        frequency: Frequency | None = None,
        values: Any | None = None,
        func: Callable | None = None,
        populate: bool = True,
    ) -> None:
        """
················································································

==Create a new `Series` object==

```
self = Series(
    start=start,
    values=values,
)
```

```
self = Series(
    periods=periods,
    values=values,
)
```

```
self = Series(
    periods=periods,
    func=func,
)
```


### Input arguments ###

???+ input "start"
    The time [`Period`](periods.md) of the first value in the `values`.

???+ input "periods"
    An iterable of time [`Periods`](periods.md) that will be used to time stamp
    the `values`. The iterable can be e.g. a tuple, a list, a time
    [`Span`](spans.md), or a single time [`Period`](periods.md).

???+ input "values"
    Time series values, supplied either as a single values, a tuple of values,
    or a NumPy array.

???+ input "func"
    A function that will be used to populate the time series; the function
    should not take any input arguments, and should return a single (scalar)
    numerical value; the function will called once for each period and each
    variant.


### Returns ###

???+ returns "None"
    This method modifies `self` in-place and does not return a value.


················································································
        """
        self.start = None
        self.data_type = data_type
        self.data = _np.full((0, num_variants), _np.nan, dtype=self.data_type, )
        self.metadata = {}
        self.__description__ = description
        #
        start = start_date if start is None else start
        periods = dates if periods is None else periods
        #
        if populate:
            test = (x is not None for x in (start, periods, values, func))
            populator = _SERIES_POPULATOR.get(tuple(test), _invalid_constructor)
            populator(
                self,
                start=start,
                periods=periods,
                frequency=frequency,
                values=values,
                func=func,
            )

    @classmethod
    def _guaranteed(
        klass,
        start: Period,
        values: _np.ndarray,
        description: str = "",
    ) -> Self:
        """
        """
        self = klass()
        self.start = start
        self.data = values
        self.__description__ = description
        self.trim()
        return self

    def reset(self, /, ) -> None:
        self.__init__(
            num_variants=self.num_variants,
            data_type=self.data_type,
        )

    def _create_periods_of_missing_values(self, num_rows=0) -> _np.ndarray:
        """
        """
        return _np.full(
            (num_rows, self.shape[1], ),
            _np.nan,
            dtype=self.data_type
        )

    @property
    @_dm.reference(category="property", )
    def shape(self, /, ) -> tuple[int, int]:
        """==Shape of time series data=="""
        return self.data.shape

    @property
    @_dm.reference(category="property", )
    def num_periods(self, /, ) -> int:
        """==Number of periods from the first to the last observation=="""
        return self.data.shape[0]

    @property
    @_dm.reference(category="property", )
    def num_variants(self, /, ) -> int:
        """==Number of variants (columns) within the `Series` object=="""
        return self.data.shape[1]

    @property
    def is_singleton(self, /, ) -> bool:
        """
        True for time series with only one variant
        """
        return _has_variants.is_singleton(self.num_variants, )

    @property
    @_dm.reference(category="property", )
    def span(self, ):
        """==Time span of the time series=="""
        return Span(self.start, self.end, ) if self.start else ()

    range = span

    @property
    @_dm.reference(category="property", )
    def from_until(self, ):
        """==Two-tuple with the start date and end date of the time series=="""
        return self.start, self.end

    @property
    @_dm.reference(category="property", )
    def periods(self, /, ) -> tuple[Period, ...]:
        """==N-tuple with the periods from the start period to the end period of the time series=="""
        return tuple(self.range, )

    dates = periods

    @property
    @_dm.reference(category="property", call_name="start", )
    def _start(self):
        """==Start date of the time series=="""
        raise NotImplementedError

    @property
    def start_period(self, /, ):
        return self.start

    start_date = start_period

    @property
    @_dm.reference(category="property", )
    def end(self):
        """==End period of the time series=="""
        return (
            self.start + self.data.shape[0] - 1
            if self.start else None
        )

    end_period = end
    end_date = end

    @property
    @_dm.reference(category="property", )
    def frequency(self, /, ):
        """==Date frequency of the time series=="""
        return (
            self.start.frequency
            if self.start is not None
            else Frequency.UNKNOWN
        )

    @property
    @_dm.reference(category="property", )
    def is_empty(self, ) -> bool:
        """==True if the time series is empty=="""
        return not self.data.size

    @property
    @_dm.reference(category="property", )
    def has_missing(self, /, ):
        """==True if the time series is non-empty and contains in-sample missing values=="""
        return (not self.is_empty) and _np.isnan(self.data).any()

    def any_missing(self, *args, ) -> bool:
        """
        """
        return self._func_missing(_np.any, *args, )

    def all_missing(self, *args, ) -> bool:
        """
        """
        return self._func_missing(_np.all, *args, )

    def count_missing(self, *args, ) -> int:
        """
        """
        return self._func_missing(_np.count_nonzero, *args, )

    def _func_missing(self, func, *args, ) -> bool:
        """
        """
        data = self.get_data(*args, )
        return func(_np.isnan(data))

    def set_data(
        self,
        dates: Dates,
        data: Any | Series,
        variants: VariantsRequestType = None,
        /,
    ) -> None:
        """
        """
        dates = self._resolve_dates(dates, )
        is_data_ndarray = isinstance(data, _np.ndarray)
        is_empty_data = (
            data is None
            or (is_data_ndarray and data.size == 0)
        )
        if not dates and is_empty_data:
            return
        data = (
            _reshape_numpy_array(data, )
            if is_data_ndarray else data
        )
        vids = self._resolve_variants(variants, )
        if not self.start:
            self.start = next(iter(dates), None, )
            self.data = self._create_periods_of_missing_values(num_rows=1, )
        pos, add_before, add_after = _get_date_positions(dates, self.start, self.shape[0], )
        self.data = self._create_expanded_data(add_before, add_after)
        if add_before:
            self.start -= add_before
        if hasattr(data, "get_data"):
            data = data.get_data(dates)
        #
        data_variants = _has_variants.iter_variants(data, )
        for c, d in zip(vids, data_variants, ):
            self.data[pos, c] = d
        self.trim()

    def _get_data_and_recreate(
        self,
        *args,
    ) -> _np.ndarray:
        """
        """
        dates, pos, variants, expanded_data = self._resolve_dates_and_positions(*args, )
        data = expanded_data[_np.ix_(pos, variants)]
        num_variants = data.shape[1]
        new = Series(num_variants=num_variants, data_type=self.data_type)
        new.set_data(dates, data)
        return new

    def _resolve_dates_and_positions(
        self,
        dates: Dates,
        variants: VariantsRequestType = None,
        /,
    ) -> tuple[Iterable[Period], Iterable[int], Iterable[int], _np.ndarray]:
        """
        """
        dates = [ t for t in self._resolve_dates(dates, ) ]
        variants = self._resolve_variants(variants)
        if not dates:
            dates = ()
            pos = ()
            data = self._create_periods_of_missing_values(num_rows=0, )[:, variants]
            return dates, pos, variants, data
        #
        base_date = self.start or min(dates, )
        pos, add_before, add_after = _get_date_positions(dates, base_date, self.shape[0], )
        data = self._create_expanded_data(add_before, add_after, )
        if not isinstance(pos, Iterable):
            pos = (pos, )
        return dates, pos, variants, data

    def alter_num_variants(
        self,
        new_num: int,
        /,
    ) -> Self:
        """
        Alter (expand, shrink) the number of variants in this time series object
        """
        if new_num < self.num_variants:
            self.shrink_num_variants(new_num, )
        elif new_num > self.num_variants:
            self.expand_num_variants(new_num, )

    def expand_num_variants(
        self,
        new_num: int,
        /,
    ) -> None:
        """
        """
        add = new_num - self.num_variants
        if add < 0:
            raise ValueError("Use shrink_num_variants to shrink the number of variants")
        data = _np.tile(self.data[:, (-1, )], (1, add, ))
        self.data = _np.hstack((self.data, data, ), )

    def shrink_num_variants(
        self,
        new_num: int,
        /,
    ) -> None:
        """
        """
        remove = self.num_variants - new_num
        if remove < 0:
            raise ValueError("Use expand_num_variants to expand the number of variants")
        self.data = self.data[:, :-remove]

    def get_values(
        self,
        *args,
        unpack_singleton: bool = True,
        **kwargs,
    ) -> _np.ndarray:
        """
        """
        data = self.get_data(*args, **kwargs, )
        values = [ tuple(data_column.tolist()) for data_column in data.T ]
        return _has_variants.unpack_singleton(
            values, self.is_singleton,
            unpack_singleton=unpack_singleton,
        )

    def get_data(self, *args, **kwargs, ) -> _np.ndarray:
        return self.get_data_and_periods(*args, **kwargs, )[0]

    def get_data_and_periods(
        self,
        dates: Iterable[Period] | None | EllipsisType = ...,
        *args,
    ) -> _np.ndarray:
        periods, pos, variants, expanded_values = self._resolve_dates_and_positions(dates, *args, )
        return expanded_values[_np.ix_(pos, variants)], periods

    def get_data_variant(
        self,
        dates: Dates,
        variant: Real | None = None,
        /,
    ) -> _np.ndarray:
        """
        """
        variant = variant if variant and variant<self.data.shape[1] else 0
        return self.get_data(dates, variant, )

    def get_data_from_until(
        self,
        from_until,
        *args,
    ) -> _np.ndarray:
        """
        """
        _, pos, variants, expanded_data \
            = self._resolve_dates_and_positions(from_until, *args, )
        from_pos, to_pos = pos[0], pos[-1]+1
        return expanded_data[from_pos:to_pos, variants]

    def get_data_variant_from_until(
        self,
        from_until: Iterable[Period],
        variant: int | None = None,
        /,
    ) -> _np.ndarray:
        """
        """
        variant = variant if variant and variant < self.data.shape[1] else 0
        return self.get_data_from_until(from_until, variant, )

    def extract_variants(
        self,
        variants,
        /,
    ) -> None:
        if not isinstance(variants, Iterable):
            variants = (variants, )
        else:
            variants = tuple(c for c in variants)
        self.data = self.data[:, variants]

    def set_start(
        self,
        new_start: Period,
        /,
    ) -> Self:
        self.start = new_start
        return self

    def _resolve_dates(self, dates, ) -> tuple[Period]:
        """
        """
        if dates is None:
            # dates = []
            dates = ...
        if isinstance(dates, slice) and dates == slice(None, ):
            dates = ...
        if dates is ... and self.start is not None:
            dates = Span(None, None, )
        if dates is ... and self.start is None:
            dates = EmptyRanger()
        if hasattr(dates, "needs_resolve") and dates.needs_resolve:
            dates = dates.resolve(self, )
        return tuple(
            d.resolve(self) if hasattr(d, "needs_resolve") and d.needs_resolve else d
            for d in dates
        )

    def _resolve_variants(
        self,
        variants: VariantsRequestType,
        /,
    ) -> Iterable[int]:
        """
        Resolve variant request to an iterable of integers
        """
        if variants is None or variants is Ellipsis:
            variants = slice(None, )
        if isinstance(variants, slice):
            variants = range(*variants.indices(self.num_variants))
        if not isinstance(variants, Iterable):
            variants = (variants, )
        return variants

    def shift(
        self,
        by: ShiftType = -1,
        **kwargs,
    ) -> None:
        r"""
................................................................................

==Shift the time series start date==

Shift the start date of the time series by a number of periods or to a specific
date.

    self.shift(by=-1, )

### Input arguments ###

???+ input "self"
    The current time series object that will be shifted.

???+ input "by"
    The number of periods to shift the observations by. If `by` is a string,
    the observations are manipulated as follows:

    * `"yoy"`: Shift all observations by one year back.

    * `"soy"`: Shift to the start of the year.

    * `"eopy"`: Shift to the end of the previous year.

................................................................................
        """
        if isinstance(by, int):
            self._shift_by_number(by, **kwargs, )
        else:
            method_name = f"_shift_{by}"
            getattr(self, method_name)(**kwargs, )

    def _shift_by_number(self, by: int, **kwargs, ) -> None:
        r"""
        Shift (lag, lead) start date by a number of periods
        """
        if self.start is None:
            return
        self.start -= by

    def _shift_yoy(self, /, **kwargs, ) -> None:
        r"""
        Shift the start date by one year back
        """
        self._shift_by_number(-self.frequency.value, )

    def _shift_soy(self, /, **kwargs, ) -> None:
        r"""
        Replace each observation by the start of the year observation
        """
        new_data = self.get_data(
            t.create_soy()
            for t in self.range
        )
        self._replace_data(new_data, )

    def _shift_eopy(self, /, **kwargs, ) -> Self:
        r"""
        Replace each observation by the end of the previous year observation
        """
        new_data = self.get_data(
            t.create_eopy()
            for t in self.range
        )
        self._replace_data(new_data, )

    def _shift_tty(
        self,
        by: Any | None = None,
        *,
        neutral_value: int | None = None,
        **kwargs,
    ) -> None:
        r"""
        Replace each observation by the start of the year observation except for
        the start-of-year periods which are filled with a neutral value.
        """
        zipped = tuple((t, t.create_tty(), ) for t in self.span)
        neutral_periods = tuple(t for t, tty in zipped if tty is None)
        periods = (t for t, tty in zipped if tty is not None )
        dty_periods = (tty for _, tty in zipped if tty is not None)
        dty_values = self.get_data(dty_periods, )
        self.set_data(periods, dty_values, )
        self.set_data(neutral_periods, neutral_value, )

    def hstack(self, *args, ):
        """
        """
        if not args:
            return self.copy()
        self_args = (self, *args, )
        if all(i.is_empty for i in self_args):
            num_variants = sum(i.num_variants for i in self_args)
            return Series(num_variants=num_variants, )
        encompassing_span, *from_until = _dates.get_encompassing_span(self, *args, )
        new_data = self.get_data_from_until(from_until, )
        add_data = (
            x.get_data_from_until(from_until, )
            if hasattr(x, "get_data")
            else _create_data_variant_from_number(x, encompassing_span, self.data_type)
            for x in args
        )
        new_data = _np.hstack((new_data, *add_data))
        new = Series(num_variants=new_data.shape[1], )
        new.set_data(encompassing_span, new_data, )
        return new

    @_dm.reference(category="homogenizing", )
    def clip(
        self,
        /,
        new_start: Period | None,
        new_end: Period | None,
    ) -> None:
        r"""
················································································

==Clip time series to a new start and end period==

    self.clip(new_start, new_end)


### Input arguments ###


???+ input "new_start"
    The new start period for the `self` time series; if `None`, the current
    start period is kept.

???+ input "new_end"
    The new end period for the `self` time series; if `None`, the current
    end period is kept.


### Returns ###


???+ returns "None"
    This method modifies `self` in-place and does not return a value.


················································································
        """
        if new_start is None or new_start < self.start:
            new_start = self.start
        if new_end is None or new_end > self.end:
            new_end = self.end
        if new_start == self.start and new_end == self.end:
            return
        self.data = self.get_data_from_until((new_start, new_end, ), )
        self.start = new_start

    def empty(
        self,
        *,
        num_variants: int | None = None
    ) -> None:
        num_variants = num_variants if num_variants is not None else self.num_variants
        self.data = _np.empty((0, num_variants), dtype=self.data_type)

    @_dm.no_reference
    def overlay_by_span(
        self,
        other: Self,
        /,
    ) -> None:
        self.set_data(other.range, other.data, )
        self.trim()

    @_dm.reference(category="multiple", )
    def overlay(
        self,
        other: Self,
        /,
        method: LayMethodType = "by_span",
    ) -> None:
        r"""
................................................................................

==Overlay another time series values onto the current time series==

Overlay the values of another time series onto the current time series on the
entire span of the other time series, i.e. from the start to the end period
regardless of missing in-sample values.

    self.overlay(
        other,
        method="by_span",
    )

### Input arguments ###

???+ input "self"
    The current time series object.

???+ input "other"
    The time series object whose values will be overlaid onto the current time
    series.

???+ input "method"
    The method to use for overlaying the values. The default (and currently the
    only available) method is `"by_span"`.

### Returns ###

This method modifies `self` in place and returns `None`.

### Details ###

???+ abstract "Algorithm"

    The resulting time series is determined the following way:

    1. The span of the resulting series starts at the earliest start period of the two
    series and ends at the latest end period of the two series.

    2. The observations from the `self` (current) time series used to fill the
    resulting time span.

    3. Within the span of the `other` time series (from the first available
    observation to the last available observation), the observations from this
    `other` time series are superimposed on the resulting time series, including any
    in-sample missing observations.

................................................................................
"""
        _broadcast_variants_if_needed(self, other, )
        getattr(type(self), "overlay_" + method)(self, other, )

    @_dm.no_reference
    def underlay_by_span(
        self,
        other: Self,
        /,
    ) -> None:
        r"""
        """
        new_self = other.copy()
        new_self.overlay(self, )
        self._shallow_copy_data(new_self, )

    @_dm.reference(category="multiple", )
    def underlay(
        self,
        other: Self,
        /,
        method: LayMethodType = "by_span",
    ) -> None:
        r"""
................................................................................

==Underlay another time series values beneath the current time series==

Underlay the values of another time series beneath the current time series on
the entire span of the other time series, i.e. from the start to the end period
regardless of missing in-sample values.

    self.underlay(
        other,
        method="by_span",
    )

### Input arguments ###

???+ input "self"
    The current time series object.

???+ input "other"
    The time series object whose values will be underlaid beneath the current
    time series.

???+ input "method"
    The method to use for underlaying the values. The default (and currently the
    only available) method is `"by_span"`.

### Returns ###

This method modifies `self` in place and returns `None`.

### Details ###

???+ abstract "Algorithm"

    The resulting time series is determined the following way:

    1. The span of the resulting series starts at the earliest start period of the two
    series and ends at the latest end period of the two series.

    2. The observations from the `other` time series used to fill the
    resulting time span.

    3. Within the span of the `self` time series (from the first available
    observation to the last available observation), the observations from this
    `self` time series are superimposed on the resulting time series, including any
    in-sample missing observations.

................................................................................
        """
        _broadcast_variants_if_needed(self, other, )
        getattr(type(self), "underlay_" + method)(self, other, )

    def redate(
        self,
        new_date: Period,
        old_date: Period | None = None,
    ) -> None:
        """
        """
        if old_data is None:
            self.start = new_date
        else:
            self.start = new_date - (old_date - self.start)

    @_dm.reference(category="manipulation", )
    def replace_where(
        self,
        test: Callable,
        new_value: Real,
    ) -> None:
        """
················································································

==Replace time series values that pass a test==

```
self.replace_where(
    test,
    new_value,
)
```


### Input arguments ###


???+ input "self"
    Time series whose observations will be tested and those passing the test
    replaced.

???+ input "test"
    A function (or a Callable) that takes a numpy array and returns `True` or
    `False` for each individual value.

???+ input "new_value"
    The value to replace the observations that pass the test.


### Returns ###


???+ returns "None"
    This method modifies `self` in-place and does not return a value.


················································································
        """
        self.data[test(self.data)] = new_value
        self.trim()

    def _shallow_copy_data(
        self,
        other: Self,
        /,
    ) -> None:
        """
        """
        for n in ("start", "data", "data_type", ):
            setattr(self, n, getattr(other, n, ))

    def __and__(self, other):
        """
        Implement the & operator as hstack
        """
        return self.hstack(other, )

    __or__ = __and__

    @_dm.reference(category="manipulation", )
    def trim(self, /, ) -> None:
        r"""
................................................................................

==Trim time series data==

Trim leading and trailing missing values from the time series data.

    self.trim()

### Input arguments ###

???+ input "self"
    The time series object to trim.

### Returns ###

This method modifies `self` in place and returns `None`.

................................................................................
        """
        if self.data.size == 0:
            self.reset()
            return self
        num_leading, num_trailing \
            = _get_num_leading_trailing_missing_rows(self.data, )
        if num_trailing == self.data.shape[0]:
            self.reset()
            return self
        if not num_leading and not num_trailing:
            return self
        slice_from = num_leading or None
        slice_to = -num_trailing if num_trailing else None
        self.data = self.data[slice_from:slice_to, ...]
        if slice_from:
            self.start += int(slice_from)
        return self

    def _create_expanded_data(self, add_before, add_after):
        return _np.pad(
            self.data, ((add_before, add_after), (0, 0)),
            mode="constant", constant_values=_np.nan,
        )

    def _check_data_shape(self, data, /, ):
        if data.shape[1] != self.data.shape[1]:
            raise Exception("Time series data being assigned must preserve the number of variants")

    def __bool__(self, /, ):
        """
        """
        return self.data.size > 0

    def __neg__(self):
        """
        -self
        """
        output = self.copy()
        output.data = -output.data
        return output

    def __pos__(self):
        """
        +self
        """
        return self.copy()

    def __add__(self, other, ) -> Self | _np.ndarray | Real:
        """
        self + other
        """
        # arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, _op.add, )

    def __radd__(self, other):
        """
        other + self
        """
        return self.apply(lambda data: data.__radd__(other))

    def __mul__(self, other) -> Self|_np.ndarray:
        """
        self + other
        """
        # arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, _op.mul, )

    def __rmul__(self, other):
        """
        other + self
        """
        return self.apply(lambda data: data.__rmul__(other))

    def __sub__(self, other):
        """
        self - other
        """
        # arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, _op.sub, )

    def __rsub__(self, other):
        """
        other - self
        """
        return self.apply(lambda data: data.__rsub__(other))

    def __pow__(self, other):
        """
        self ** other
        """
        # FIXME: 1 ** numpy.nan -> 1 !!!!!
        # arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, _op.pow, )

    def __rpow__(self, other):
        """
        other ** self
        """
        return self.apply(lambda data: data.__rpow__(other))

    def __truediv__(self, other):
        """
        self - other
        """
        # arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, _op.truediv, )

    def __rtruediv__(self, other):
        """
        other - self
        """
        return self.apply(lambda data: data.__rtruediv__(other))

    def __floordiv__(self, other):
        """
        self // other
        """
        return self._binop(other, _op.floordiv, )

    def __rfloordiv__(self, other):
        """
        other // self
        """
        return self.apply(lambda data: data.__rfloordiv__(other))

    def __mod__(self, other):
        """
        self % other
        """
        return self._binop(other, _op.mod, )

    def __rmod__(self, other, /, ):
        """
        other % self
        """
        return self.apply(lambda data: data.__rmod__(other))

    def __round__(self, *args, **kwargs, ):
        """
        """
        new = self.copy()
        new.data = _np.round(new.data, *args, **kwargs, )
        return new

    for n in ["gt", "lt", "ge", "le", "eq", "ne", ]:
        exec(f"def __{n}__(self, other): return self._binop(other, _op.{n}, )", )

    def apply(self, func, /, *args, **kwargs, ):
        new_data = func(self.data, *args, **kwargs, )
        axis = kwargs.get("axis", None, )
        if new_data.shape == self.data.shape:
            new = self.copy()
            new._replace_data(new_data, )
            return new
        elif (axis is None or axis == 1) and new_data.shape == (self.data.shape[0], ):
            new = self.copy()
            new.data = new_data.reshape(self.data.shape[0], 1, )
            new.trim()
            return new
        elif axis is None and new_data.shape == ():
            return new_data
        elif axis == 0 and new_data.shape == (self.data.shape[1], ):
            return new_data
        else:
            raise _wrongdoings.IrisPieError(
                "Function applied on a time series resulted"
                " in a data array with an unexpected shape"
            )

    def _binop(self, other, func, /, new=None, ):
        if not isinstance(other, type(self)):
            return self.apply(lambda data: func(data, other))
        # FIXME: empty encompassing range
        _, *from_until = _dates.get_encompassing_span(self, other)
        self_data = self.get_data_from_until(from_until, )
        other_data = other.get_data_from_until(from_until, )
        new_data = func(self_data, other_data)
        new = Series(num_variants=new_data.shape[1], ) if new is None else new
        new._replace_start_and_values(from_until[0], new_data, )
        return new

    def _broadcast_variants(self, num_variants, /, ) -> None:
        """
        """
        if self.data.shape[1] == num_variants:
            return
        if self.data.shape[1] == 1:
            self.data = _np.repeat(self.data, num_variants, axis=1, )
            return
        raise _wrongdoings.IrisPieError("Cannot broadcast variants")

    def _replace_data(
        self,
        new_values,
        /,
    ) -> None:
        """
        """
        self.data = new_values
        self.trim()

    def _replace_start_and_values(
        self,
        new_start,
        new_values,
        /,
    ) -> None:
        """
        """
        self.start = new_start
        self._replace_data(new_values, )

    def iter_dates_values(self, /, unpack_singleton=True, ):
        """
        Default iterator is line by line, yielding a tuple of (date, values)
        """
        def _unpack_singleton_data_row(data_list: list[Real], /, ):
            return _has_variants.unpack_singleton(data_list, True, )
        def _keep_data_row(data_list: list[Real], /, ):
            return data_list
        data_row_func = (
            _unpack_singleton_data_row
            if self.is_singleton and unpack_singleton
            else _keep_data_row
        )
        for date, data_row in zip(self.range, self.data, ):
            yield date, data_row_func(data_row.tolist(), )

    def iter_variants(self, /, ) -> Iterator[Self]:
        """
        """
        for data in self.iter_data_variants_from_until(..., ):
            new = self.copy()
            new._replace_data(data, )
            yield new

    def iter_own_data_variants_from_until(self, from_until, ) -> Iterator[_np.ndarray]:
        """
        Iterates over the data variants from the given start date to the given end date
        """
        data_from_until = self.data if from_until == ... else self.get_data_from_until(from_until, )
        return iter(data_from_until.T, )

    def iter_data_variants_from_until(self, from_until, ) -> Iterator[_np.ndarray]:
        """
        Iterates over the data variants from the given start date to the given end date
        """
        return _iterators.exhaust_then_last(self.iter_own_data_variants_from_until(from_until, ), )

    def logistic(self, /, ) -> Self:
        """
        """
        self.data = 1 / (1 + _np.exp(-self.data))

    #]


def _get_num_leading_trailing_missing_rows(data: _np.ndarray, /, ):
    """
    """
    #[
    # Boolean index of rows with at least one observation
    boolex_observations = ~_np.all(_np.isnan(data, ), axis=1, )
    if _np.all(boolex_observations, ):
        # All rows have at least one observation
        num_leading = 0
        num_trailing = 0
    elif not _np.any(boolex_observations, ):
        # No row has any observation
        num_leading = 0
        num_trailing = data.shape[0]
    else:
        # Some rows have observations
        num_leading = _np.argmax(boolex_observations)
        num_trailing = _np.argmax(boolex_observations[::-1])
    return num_leading, num_trailing
    #]


def hstack(first, *args) -> Self:
    return first.hstack(*args)


def _create_data_variant_from_number(
    number: Real,
    span: Span,
    data_type: type,
    /,
) -> _np.ndarray:
    return _np.full((len(span), 1), number, dtype=data_type)


def _from_periods_and_values(
    self,
    periods: Iterable[Period] | str,
    values: _np.ndarray | Iterable,
    frequency: Frequency | None = None,
    **kwargs,
) -> None:
    """
    """
    #[
    # dates = _dates.ensure_period_tuple(dates, frequency=frequency, )
    self.set_data(periods, values, )
    #]


def _from_periods_and_func(
    self,
    periods: Iterable[Period] | str,
    func: Callable,
    frequency: Frequency | None = None,
    **kwargs,
) -> Self:
    """
    Create a new time series from dates and a function
    """
    #[
    # dates = _dates.ensure_period_tuple(dates, frequency=frequency, )
    data = [
        [func() for j in range(self.num_variants)]
        for i in range(len(periods))
    ]
    data = _np.array(data, dtype=self.data_type)
    self.set_data(periods, data)
    #]


def _from_start_and_values(
    self,
    start: Period,
    values: _np.ndarray | Iterable,
    frequency: Frequency | None = None,
    **kwargs,
) -> None:
    """
    """
    #[
    # start = _dates.ensure_period_tuple(start, frequency=frequency, )[0]
    self.start = start
    if isinstance(values, _np.ndarray):
        values = _reshape_numpy_array(values, )
    else:
        values = _has_variants.iter_variants(values, )
        values = _np.column_stack(tuple(
            v for v, _ in zip(values, range(self.num_variants), )
        ))
    values = values.astype(self.data_type, )
    self.data = values
    self.trim()
    #]


def _reshape_numpy_array(values: _np.ndarray, /, ) -> _np.ndarray:
    """
    """
    #[
    return (
        values.reshape(values.shape[0], -1)
        if values.ndim >= 2 else values.reshape(-1, 1)
    )
    #]


def _broadcast_variants_if_needed(
    self: Series,
    other: Series,
    /,
) -> tuple[Series, Series]:
    """
    """
    #[
    if self.num_variants == other.num_variants:
        return self, other
    #
    if self.num_variants == 1:
        return self._broadcast_variants(other.num_variants, ), other
    #
    if other.num_variants == 1:
        return self, other._broadcast_variants(self.num_variants, )
    #
    raise _wrongdoings.IrisPieError("Cannot broadcast time series variants")
    #]


def _invalid_constructor(
    self,
    **kwargs,
) -> NoReturn:
    raise _wrongdoings.IrisPieError("Invalid Series object constructor")


_SERIES_POPULATOR = {
    (False, False, False, False): lambda self, **kwargs: None,
    (True, False, True, False): _from_start_and_values,
    (False, True, True, False): _from_periods_and_values,
    (False, True, False, True): _from_periods_and_func,
}


for n in ("shift", "redate", "underlay", "overlay", ):
    code = FUNC_STRING.format(n=n, )
    exec(code, globals(), locals(), )
    __all__.append(n)





"""
Facades for steady-state solvers
"""


#[

from __future__ import annotations

import numpy as _np
import scipy as _sp
import neqs as _ne

from .evaluators import SteadyEvaluator

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    SolverType = Literal["neqs_levenberg", "neqs_firefly", "scipy_root", ]

#]


#===============================================================================


_DEFAULT_SCIPY_ROOT_SETTINGS = {
    "method": "lm",
    "tol": 1e-12,
}


def scipy_root(
    steady_evaluator: SteadyEquator,
    maybelog_init_guess: _np.ndarray,
    *,
    solver_settings: dict[str, Any] | None = None,
) -> tuple[bool, _np.ndarray, Any, ]:
    """
    """
    solver_settings = solver_settings or {}
    solver_settings = _DEFAULT_SCIPY_ROOT_SETTINGS | solver_settings
    #
    root_final = _sp.optimize.root(
        steady_evaluator.eval,
        maybelog_init_guess,
        jac=True,
        **solver_settings,
    )
    #
    steady_evaluator.iter_printer.print_footer()
    maybelog_final_guess = root_final.x
    func_norm = _sp.linalg.norm(root_final.fun, 2, )
    success = root_final.success and func_norm < solver_settings["tol"]
    # TODO: create exit status based on root_final
    exit_status = None
    return maybelog_final_guess, success, exit_status,


#===============================================================================


_DEFAULT_NEQS_LEVENBERG_SETTINGS = {
    "step_tolerance": float("inf"),
    "func_tolerance": 1e-14,
}


def neqs_levenberg(
    steady_evaluator: SteadyEquator,
    maybelog_init_guess: _np.ndarray,
    *,
    solver_settings: dict[str, Any] | None = None,
) -> tuple[bool, _np.ndarray, Any, ]:
    """
    """
    solver_settings = solver_settings or {}
    solver_settings = _DEFAULT_NEQS_LEVENBERG_SETTINGS | solver_settings
    maybelog_final_guess, exit_status = _ne.levenberg(
        eval_func=steady_evaluator.eval_func,
        eval_jacob=steady_evaluator.eval_jacob,
        init_guess=maybelog_init_guess,
        iter_printer=steady_evaluator.iter_printer,
        **solver_settings,
    )
    success = exit_status.is_success
    return maybelog_final_guess, success, exit_status,


#===============================================================================


_DEFAULT_NEQS_FIREFLY_SETTINGS = {
}


def neqs_firefly(
    steady_evaluator: SteadyEquator,
    maybelog_init_guess: _np.ndarray,
    *,
    solver_settings: dict[str, Any] | None = None,
) -> tuple[bool, _np.ndarray, Any, ]:
    """
    """
    solver_settings = solver_settings or {}
    solver_settings = _DEFAULT_NEQS_FIREFLY_SETTINGS | solver_settings
    #
    maybelog_final_guess = _ne.firefly(
        eval_func=steady_evaluator.eval_func,
        init_guess=maybelog_init_guess,
        iter_printer=steady_evaluator.iter_printer,
        solver_settings=solver_settings,
    )
    #
    success = True
    exit_status = None
    return maybelog_final_guess, success, exit_status,





"""
Time series indexing inlay
"""


#[

from __future__ import annotations

import documark as _dm

from .. import dates as _dates
from ._categories import CATEGORIES

#]


class Inlay:
    """
    """
    #[

    @_dm.reference(
        category=None,
        call_name=CATEGORIES["indexing"],
        call_name_is_code=False,
        priority=30,
    )
    def indexing(self, /, ) -> None:
        r"""
················································································

Time `Series` objects can be indexed in four ways (note the square versus round
brackets):

| Indexing                                           | Description
|----------------------------------------------------|-------------
| `self[shift]`                                      | Time shift
| `self[dates]`, `self[dates, variants]`             | Data extraction
| `self[dates] = ...`, `self[dates, variants] = ...` | Data assignment
| `self(dates)`, `self(dates, variants)`             | Time `Series` recreation


### Time shift ###

```
self[shift]
```

Time shift is done by passing an integer to the `self[shift]` or
`self[shift]` indexing. The time shift syntax returns a new copy of the
original series, with the time periods shifted by `shift`.


### Data extractation ###

```
self[dates]
self[dates, variants]
```

The `dates` is a `Dater` or a tuple of `Daters` or a time `Span` object,
and `variants` is an integer or a tuple of integers or a `slice` object
specifying the variants. The data extraction syntax returns a
two-dimensional `numpy` array, with the time dimension running along the
rows and the variant dimension running along the columns.


### Data assignment ###

```
self[dates] = ...
self[dates, variants] = ...
```

The `dates` is a `Dater` or a tuple of `Daters` or a time `Span` object,
and `variants` is an integer or a tuple of integers or a `slice` object
specifying the variants. The data assignment syntax sets the data in the
time series.


### Time `Series` recreation ###

```
self(dates)
self(dates, variants)
```

The `dates` is a `Dater` or a tuple of `Daters` or a time `Span` object,
and `variants` is an integer or a tuple of integers or a `slice` object
specifying the variants. The time `Series` recreation syntax returns a new
time `Series` object based on the data selected by the `dates` and
`variants`.

················································································
        """
        raise NotImplementedError

    def __getitem__(
        self,
        index: int | tuple,
        /,
    ) -> _np.ndarray:
        """
        Get data self[dates] or self[dates, variants]
        """
        if isinstance(index, int, ):
            # Time shift
            new = self.copy()
            new.shift(index, )
            return new
        else:
            # Extract data
            if isinstance(index, tuple, ):
                dates, variants = index
            else:
                dates, variants = index, None
            # dates = _dates.ensure_period_tuple(dates, frequency=self.frequency, )
            return self.get_data(dates, variants, )

    def __setitem__(
        self,
        index: int | tuple,
        data,
        /,
    ) -> None:
        """
        Set data self[dates] = ... or self[dates, variants] = ...
        """
        if not isinstance(index, tuple):
            index = (index, None, )
        return self.set_data(index[0], data, index[1], )

    def __call__(
        self,
        *index,
    ) -> Self:
        """
        Create a new time series based on date retrieved by self[dates] or self[dates, variants]
        """
        return self._get_data_and_recreate(*index, )

    #]





import irispie as ir
import irispie.series.functions as isf
import numpy as _np
import scipy as _sp
from irispie.series._elementwise import _ONE_ARG_FUNCTION_DISPATCH, _TWO_ARGS_FUNCTION_DISPATCH

import random as rn
import pytest


span = ir.qq(2021,1) >> ir.qq(2025,4)

x = ir.Series(periods=span, num_variants=2, func=lambda: rn.uniform(0.1, 0.9))


parameter_names = ("func_name", )
parameter_values = [ (n, ) for n in _ONE_ARG_FUNCTION_DISPATCH.keys() ]
@pytest.mark.parametrize(parameter_names, parameter_values, )
def test_elementwise_one_arg(func_name, ):
    y1 = x.copy()
    getattr(y1, func_name)()
    #
    y2 = getattr(isf, func_name)(x, )
    #
    y3 = getattr(ir, func_name)(x, )
    #
    implementation_name = _ONE_ARG_FUNCTION_DISPATCH[func_name]
    expected_data = eval(f"{implementation_name}(x.data, )", )
    #
    assert _np.all(y1.data == expected_data)
    assert _np.all(y2.data == expected_data)
    assert _np.all(y3.data == expected_data)


parameter_names = ("func_name", )
parameter_values = [ (n, ) for n in _TWO_ARGS_FUNCTION_DISPATCH.keys() ]
@pytest.mark.parametrize(parameter_names, parameter_values, )
def test_elementwise_two_args(func_name, ):
    y1 = x.copy()
    getattr(y1, func_name)(1, )
    #
    y2 = getattr(isf, func_name)(x, 1, )
    #
    y3 = getattr(ir, func_name)(x, 1, )
    #
    implementation_name = _TWO_ARGS_FUNCTION_DISPATCH[func_name]
    expected_data = eval(f"{implementation_name}(x.data, 1, )", )
    #
    assert _np.all(y1.data == expected_data)
    assert _np.all(y2.data == expected_data)
    assert _np.all(y3.data == expected_data)


if __name__ == "__main__":
    test_elementwise_one_arg("log", )
    test_elementwise_two_args("minimum", )





"""
Simultaneous model simulation inlay
"""


#[
from __future__ import annotations

import numpy as _np

from typing import Any, Literal
from .. import quantities as _quantities
from .. import has_variants as _has_variants
from .. import wrongdoings as _wrongdoings
from ..import dates as _dates
from ..databoxes.main import Databox
from ..dataslates.main import Dataslate
from ..fords import solutions as _solutions
from ..plans.simulation_plans import SimulationPlan

from ..fords import simulators as _ford_simulator
from ..period_by_period import simulators as _period_by_period_simulator
from ..stacked_time import simulators as _stacked_time_simulator

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from collections.abc import Iterable
    from ..frames import Frame
    from ..dates import Period
#]


_SIMULATOR_MODULE = {
    "first_order": _ford_simulator,
    "period_by_period": _period_by_period_simulator,
    "period": _period_by_period_simulator,
    "stacked_time": _stacked_time_simulator,
    "stacked": _stacked_time_simulator,
}


_Info = dict[str, Any] | list[dict[str, Any]]


class Inlay:
    """
    """
    #[

    def simulate(
        self,
        in_db: Databox,
        span: Iterable[Period],
        /,
        *,
        plan: SimulationPlan | None = None,
        method: Literal["first_order", "period_by_period", "period", "stacked_time", "stacked", ] = "first_order",
        prepend_input: bool = True,
        target_db: Databox | None = None,
        num_variants: int | None = None,
        remove_initial: bool = True,
        remove_terminal: bool = True,
        shocks_from_data: bool = True,
        stds_from_data: bool = True,
        parameters_from_data: bool = False,
        force_split_frames: bool = False,
        when_fails: Literal["critical", "error", "warning", "silent"] = "critical",
        #
        unpack_singleton: bool = True,
        return_info: bool = False,
        **kwargs,
    ) -> tuple[Databox, _Info]:
        """
        """

        num_variants = self.resolve_num_variants_in_context(num_variants, )

        base_dates = tuple(span, )

        if plan is not None and not plan.is_empty:
            plan.check_consistency(self, base_dates, )
            extra_databox_names = plan.get_databox_names()
        else:
            plan = None
            extra_databox_names = None

        slatable = self.get_slatable_for_simulate(
            shocks_from_data=shocks_from_data,
            stds_from_data=stds_from_data,
            parameters_from_data=parameters_from_data,
        )

        dataslate = Dataslate.from_databox_for_slatable(
            slatable, in_db, base_dates,
            num_variants=num_variants,
            extra_databox_names=extra_databox_names,
        )

        zipped = zip(
            range(num_variants, ),
            self.iter_variants(),
            dataslate.iter_variants(),
        )

        when_fails_stream = _wrongdoings.create_stream(when_fails, "Simulation failed to complete", )
        unanticipated_shock_qids = _get_unanticipated_shock_qids(self, )
        simulator_module = _SIMULATOR_MODULE[method]

        #=======================================================================
        # Main loop over variants and over frames
        #
        out_info = []
        for vid, model_v, dataslate_v in zipped:

            input_data_array = dataslate_v.get_data_variant().copy()

            frames = simulator_module.create_frames(
                model_v, dataslate_v, plan,
                force_split_frames=force_split_frames,
            )

            info_v = {
                "method": simulator_module.METHOD_NAME,
                "frames": tuple(frames),
                "exit_status": (),
                "frame_databoxes": (),
            }

            simulator_module.simulate_initial_guess(
                model_v, dataslate_v, plan,
                **kwargs,
            )

            for frame in frames:

                simulation_header = _create_simulation_header(vid, frame, )

                # Create a full copy of the dataslate, both invariant and
                # variants. This is necessary because the frame will first prune
                # the data in place, and then remove initial and terminal data
                # and periods.
                frame_ds = dataslate_v.copy()

                # Remove unanticipated data from the frame except the first
                # simulation period for SplitFrames; keep everything unchanged
                # for SingleFrames
                frame.prune_frame_data(frame_ds, unanticipated_shock_qids, )

                exit_status = simulator_module.simulate_frame(
                    model_v, frame_ds,
                    frame=frame,
                    input_data_array=input_data_array,
                    plan=plan,
                    simulation_header=simulation_header,
                    return_info=return_info,
                    **kwargs,
                )

                if not exit_status.is_success:
                    when_fails_stream.add(f"{simulation_header}: {exit_status}", )

                # In SplitFrames, copy frame data back to the main dataslate
                # * Unanticipated shocks are written back within the frame slice
                # * Other data are writen back within the simulation slice
                # In SingleFrames, do nothing because everyrhing is kept and
                # performed within the main dataslate
                frame.write_frame_data_to_main_dataslate(
                    dataslate_v, frame_ds, unanticipated_shock_qids,
                )

                # Create a frame databox only if output info is requested
                # for performance reasons
                frame_db = (
                    frame_ds.to_databox(span="base", )
                    if return_info else None
                )

                info_v["frame_databoxes"] += (frame_db, )
                info_v["exit_status"] += (exit_status, )

            out_info.append(info_v, )

        when_fails_stream._raise()
        #
        #=======================================================================
        #
        #
        # Remove initial and terminal condition data (all lags and leads
        # before and after the simulation span)
        if remove_terminal:
            dataslate.remove_terminal()
        if remove_initial:
            dataslate.remove_initial()
        #
        # Convert all variants of the dataslate to a databox
        out_db = dataslate.to_databox()
        if prepend_input:
            out_db.prepend(in_db, base_dates[0]-1, )
        #
        # Add to custom databox
        if target_db is not None:
            out_db = target_db | out_db
        #
        if return_info:
            is_singleton = num_variants == 1
            out_info = _has_variants.unpack_singleton(
                out_info, is_singleton,
                unpack_singleton=unpack_singleton,
            )
            return out_db, out_info
        else:
            return out_db
    #]


def _append_frame_databox(
    frame_databoxes: list[Databox],
    frame_ds: Dataslate,
) -> None:
    """
    """
    #[
    frame_db = frame_ds.to_databox(span="base", )
    frame_databoxes += (frame_db, )
    #]


def _get_unanticipated_shock_qids(self, /, ) -> tuple[int, ...]:
    """
    """
    #[
    plannable = self.get_simulation_plannable()
    name_to_qid = self.create_name_to_qid()
    unanticipated_shock_names = getattr(plannable, "can_be_endogenized_unanticipated", (), )
    return tuple(name_to_qid[i] for i in unanticipated_shock_names)
    #]


def _create_simulation_header(
    vid: str,
    frame: Frame,
) -> str:
    """
    """
    #[
    simulation_span = _dates.get_printable_span(frame.start, frame.simulation_end, )
    return f"[Variant {vid}][Periods {simulation_span}]"
    #]




"""
"""


#[

from __future__ import annotations

import numpy as _np
import scipy as _sp
import itertools as _it
import documark as _dm

from ..incidences import main as _incidences
from ..fords import covariances as _covariances
from .. import quantities as _quantities
from .. import namings as _namings

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from numbers import Real

#]




class Inlay:
    """
    """

    @_dm.reference(category="parameters", )
    def rescale_stds(
        self,
        factor: Real,
        *,
        kind: _quantities.QuantityKind | None = None,
    ) -> None:
        """
................................................................................

==Rescale the standard deviations of model shocks==

Adjust the standard deviations of the model shocks by a specified factor. 
This method allows scaling the standard deviations for the shocks in the 
model based on the provided factor.

    self.rescale_stds(
        factor,
        kind=None,
    )


### Input arguments ###


???+ input "factor"
    A real non-negative number by which to scale the standard deviations of the
    model shocks. This value is used to multiply the existing standard
    deviations.

???+ input "kind"
    An optional parameter to narrow down the types of shocks to rescale. It
    can be one, or a combination, of the following:

    * `ir.UNANTICIPATED_STD`
    * `ir.ANTICIPATED_STD`
    * `ir.MEASUREMENT_STD`

    If `None`, the standard deviations of all shocks will be rescaled.


### Returns ###


???+ returns "None"
    This method does not return any value but modifies the standard deviations 
    of model shocks in-place, rescaling them.

................................................................................
        """
        if 1.0 * factor <= 0:
            raise ValueError("The scaling factor must be a real non-negative number")
        std_qids = self._get_std_qids(kind=kind, )
        for v in self._variants:
            v.rescale_values("levels", factor, std_qids, )

    def get_acov(
        self,
        /,
        up_to_order: int = 0,
        return_dimension_names: bool = True,
        unpack_singleton: bool = True,
    ) -> tuple[list[_np.ndarray] | _np.ndarray, _namings.DimensionNames] | list[_np.ndarray] | _np.ndarray:
        """
        Asymptotic autocovariance of model variables
        """
        #
        # Combine vectors of transition and measurement tokens, and select
        # those with zero shift only using a boolex
        system_vector, boolex_zero_shift = _get_system_vector(self, )
        #
        # Tuple element cov_by_variant[variant_index][order] is an N-by-N
        # covariance matrix
        cov_by_variant = [
            self._getv_autocov(v, boolex_zero_shift, up_to_order=up_to_order, )
            for v in self._variants
        ]
        #
        # Tuple element cov_by_order[order] is an N-by-N covariance matrix
        # (for singleton models) or an N-by-N-by-num_variants covariance
        # matrix (for nonsingleton models)
        # stack_func = _STACK_FUNC_FACTORY[self.is_singleton]
        # cov_by_order = tuple(
        #     stack_func(tuple(cov[order] for cov in cov_by_variant))
        #     for order in range(up_to_order + 1)
        # )
        #
        cov_by_variant = self.unpack_singleton(
            cov_by_variant,
            unpack_singleton=unpack_singleton,
        )
        #
        if return_dimension_names:
            dimension_names = _get_dimension_names(self, system_vector, )
            return cov_by_variant, dimension_names
        else:
            return cov_by_variant

    def get_acorr(
        self,
        /,
        acov: tuple[_np.ndarray, ..., tuple[str], tuple[str]] | None = None,
        up_to_order: int = 0,
        unpack_singleton: bool = True,
        return_dimension_names: bool = True,
    ) -> tuple[list[_np.ndarray] | _np.ndarray, _namings.DimensionNames] | list[_np.ndarray] | _np.ndarray:
        """
        Asymptotic autocorrelation of model variables
        """
        acov_by_variant = acov
        if acov_by_variant is None:
            acov_by_variant, *_ = self.get_acov(
                up_to_order=up_to_order,
                return_dimension_names=False,
                unpack_singleton=False,
            )
        acov_by_variant = self.repack_singleton(acov_by_variant, )
        acorr_by_variant = [
            _covariances.acorr_from_acov(i)
            for i in acov_by_variant
        ]
        acorr_by_variant = self.unpack_singleton(
            acorr_by_variant,
            unpack_singleton=unpack_singleton,
        )
        if return_dimension_names:
            system_vector, *_ = _get_system_vector(self, )
            dimension_names = _get_dimension_names(self, system_vector, )
            return acorr_by_variant, dimension_names,
        else:
            return acorr_by_variant

    def _getv_autocov(
        self,
        variant: _variants.Variant,
        boolex_zero_shift: tuple[bool, ...] | Ellipsis,
        /,
        up_to_order: int = 0,
    ) -> _np.ndarray:
        """
        """
        def select(cov: _np.ndarray, /, ) -> _np.ndarray:
            """
            Select elements of solution vectors with zero shift only
            """
            cov = cov[boolex_zero_shift, :]
            cov = cov[:, boolex_zero_shift]
            return cov
        cov_u = self._getv_cov_u(variant, )
        cov_w = self._getv_cov_w(variant, )
        cov_by_order = _covariances.get_autocov_square(variant.solution, cov_u, cov_w, up_to_order, )
        return tuple(select(cov, ) for cov in cov_by_order)

    def get_stdvec_unanticipated_shocks(self, /, ):
        """
        """
        std_u = [
            self._getv_std_u(v, )
            for v in self._variants
        ]
        return self.unpack_singleton(std_u, )

    def get_stdvec_measurement_shocks(self, /, ):
        """
        """
        std_w = [
            self._getv_std_w(v, )
            for v in self._variants
        ]
        return self.unpack_singleton(std_w, )

    def get_cov_unanticipated_shocks(
        self,
        /,
        unpack_singleton: bool = True,
    ):
        """
        """
        cov_u = [
            self._getv_cov_u(v, )
            for v in self._variants
        ]
        return self.unpack_singleton(cov_u, unpack_singleton=unpack_singleton, )

    def get_cov_measurement_shocks(
        self,
        /,
        unpack_singleton: bool = True,
    ):
        """
        """
        cov_w = [
            self._getv_cov_w(v, )
            for v in self._variants
        ]
        return self.unpack_singleton(cov_w, unpack_singleton=unpack_singleton, )

    def _getv_std_u(self, variant, /, ):
        """
        """
        shocks = self._invariant.dynamic_descriptor.solution_vectors.unanticipated_shocks
        return _retrieve_stds(self, variant, shocks, )

    def _getv_std_w(self, variant, /, ):
        """
        """
        shocks = self._invariant.dynamic_descriptor.solution_vectors.measurement_shocks
        return _retrieve_stds(self, variant, shocks, )

    def _getv_cov_u(self, variant, /, ):
        """
        """
        stds = self._getv_std_u(variant, )
        return _np.diag(stds**2, )

    def _getv_cov_w(self, variant, /, ):
        """
        """
        stds = self._getv_std_w(variant, )
        return _np.diag(stds**2, )


def _stack_singleton(x: tuple[_np.ndarray], /, ) -> _np.ndarray:
    return x[0]


def _stack_nonsingleton(x: tuple[_np.ndarray], /, ) -> _np.ndarray:
    return _np.dstack(x)


_STACK_FUNC_FACTORY = {
    True: _stack_singleton,
    False: _stack_nonsingleton,
}


def _get_dimension_names(
    self,
    system_vector: tuple[_incidences.Token, ...],
) -> tuple[tuple[str, ...], tuple[str, ...]]:
    """
    """
    qid_to_name = self.create_qid_to_name()
    qid_to_logly = self.create_qid_to_logly()
    names = tuple(
        _quantities.wrap_logly(qid_to_name[tok.qid], qid_to_logly[tok.qid], )
        for tok in system_vector
    )
    return _namings.DimensionNames(rows=names, columns=names, )


def _get_system_vector(
    self,
    /,
) -> tuple[tuple[_incidences.Token, ...], tuple[bool, ...]]:
    """
    """
    system_vector = \
        self._invariant.dynamic_descriptor.solution_vectors.transition_variables \
        + self._invariant.dynamic_descriptor.solution_vectors.measurement_variables
    boolex_zero_shift = tuple(tok.shift == 0 for tok in system_vector)
    system_vector = tuple(_it.compress(system_vector, boolex_zero_shift, ))
    return system_vector, boolex_zero_shift


def _retrieve_stds(self, variant, shocks, ) -> _np.ndarray:
    """
    """
    #[
    std_qids = tuple(
        self._invariant.shock_qid_to_std_qid[t.qid]
        for t in shocks
    )
    return variant.retrieve_values_as_array("levels", std_qids, )
    #]




"""
"""


#[

from __future__ import annotations

from collections import namedtuple
import functools as _ft
import itertools as _it
import numpy as _np
import scipy as _sp
import neqs as _ne

from .. import equations as _equations
from ..incidences import blazer as _blazer
from .. import has_variants as _has_variants
from .. import wrongdoings as _wrongdoings
from ..fords import steadiers as _fs
from ..steadiers import evaluators as _evaluators
from ..steadiers import solver_dispatcher as _solver_dispatcher
from ..equations import ENDOGENOUS_EQUATION
from ..quantities import QuantityKind

from . import _flags

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..plans.steady_plans import SteadyPlan
    from ..equations import Equation
    from ._variants import Variant
    from collections.abc import Iterable, Callable
    from typing import Any, Literal, NoReturn
    from ..steadiers.solver_dispatcher import SolverType

#]


_Wrt = namedtuple("Wrt", (
    "equations",
    "eids",
    "names",
    "fixed_level_names",
    "fixed_change_names",
    "qids",
    "fixed_level_qids",
    "fixed_change_qids",
))


_STEADY_EQUATION_SOLVED = ENDOGENOUS_EQUATION

_DEFAULT_SOLVER = "neqs_levenberg"


class Inlay:
    """
    """
    #[

    def steady(
        self,
        /,
        unpack_singleton: bool = True,
        update_steady_autovalues: bool = True,
        return_info: bool = False,
        **kwargs,
    ) -> dict | list[dict]:
        """
        Calculate steady state for each Variant within this model
        """
        model_flags = self.resolve_flags(**kwargs, )
        steady_solver = self._choose_steady_solver(model_flags.is_linear, model_flags.is_flat, )
        out_info = []
        for vid, v in enumerate(self._variants, ):
            out_info_v = steady_solver(v, model_flags, vid, **kwargs, )
            out_info.append(out_info_v, )
        if update_steady_autovalues:
            self.update_steady_autovalues()
        if return_info:
            return _has_variants.unpack_singleton(
                out_info, self.is_singleton,
                unpack_singleton=unpack_singleton,
            )

    def _steady_linear(
        self,
        variant: Variant,
        model_flags: _flags.Flags,
        vid: int,
        /,
        *,
        algorithm: Callable,
        **kwargs,
    ) -> dict[str, Any]:
        """
        """
        #
        # Calculate first-order system for steady equations for this variant
        system = self._systemize(
            variant,
            self._invariant.steady_descriptor,
            model_flags,
        )
        #
        #=======================================================================
        # Core algorithm: Calculate steady state for this variant
        Xi, Y, dXi, dY = algorithm(system)
        levels = _np.hstack(( Xi.flat, Y.flat )).flatten()
        changes = _np.hstack(( dXi.flat, dY.flat )).flatten()
        #=======================================================================
        #
        # From first-order state space, extract only tokens with zero shift
        tokens = tuple(_it.chain(
            self._invariant.steady_descriptor.system_vectors.transition_variables,
            self._invariant.steady_descriptor.system_vectors.measurement_variables,
        ))
        #
        # True for tokens with zero shift, e.g. [True, False, True, ... ]
        zero_shift_index = [ not t.shift for t in tokens ]
        #
        # Extract steady levels for quantities with zero shift
        levels = levels[zero_shift_index]
        changes = changes[zero_shift_index]
        qids = tuple(t.qid for t in _it.compress(tokens, zero_shift_index))
        #
        # Delogarithmize when needed
        self.delogarithmize(levels, changes, )
        #
        variant.update_levels_from_array(levels, qids, )
        variant.update_changes_from_array(changes, qids, )
        #
        info = {}
        return info

    _steady_linear_flat = _ft.partialmethod(
        _steady_linear,
        algorithm=_fs.solve_steady_linear_flat,
    )

    _steady_linear_nonflat = _ft.partialmethod(
        _steady_linear,
        algorithm=_fs.solve_steady_linear_nonflat,
    )

    def _steady_nonlinear(
        self,
        variant: Variant,
        model_flags: _flags.Flags,
        vid: int,
        /,
        evaluator_class: type,
        *,
        plan: SteadyPlan | None = None,
        split_into_blocks: bool | None = None,
        solver: SolverType = _DEFAULT_SOLVER,
        solver_settings: dict[str, Any] | None = None,
        iter_printer_settings: dict[str, Any] | None = None,
        **kwargs,
    ) -> dict[str, Any]:
        """
        """
        iter_printer_settings = iter_printer_settings or {}
        #
        qid_to_kind = self.create_qid_to_kind()
        qid_to_name = self.create_qid_to_name()
        #
        split_into_blocks = _resolve_split_into_blocks(split_into_blocks, plan, )
        wrt = self._resolve_steady_wrt(plan, is_flat=model_flags.is_flat, )
        if split_into_blocks:
            im = _calculate_steady_incidence_matrix(wrt.equations, wrt.qids, )
            blocks = _blazer.blaze(im, wrt.eids, wrt.qids, )
        else:
            blocks = (_blazer._Block(wrt.eids, wrt.qids), )
        num_blocks = len(blocks)
        #
        all_quantities = self.get_quantities()
        all_equations = self.get_steady_equations(kind=_STEADY_EQUATION_SOLVED, )
        human_blocks \
            = _blazer.human_blocks_from_blocks \
            (blocks, all_equations, all_quantities, )
        #
        info_v = {
            "success": None,
            "blocks": [],
        }
        #=======================================================================
        for bid, (block, human_block) in enumerate(zip(blocks, human_blocks, ), ):
            #
            info_b = {
                "success": None,
                "exit_status": None,
                "equations": human_block.equations,
                "quantities": human_block.quantities,
            }
            custom_header = f"[Variant {vid}][Block {bid}]"
            #
            block_level_qids = tuple(sorted(set(block.qids) - set(wrt.fixed_level_qids)))
            block_change_qids = tuple(sorted(set(block.qids) - set(wrt.fixed_change_qids)))
            block_equations = tuple(wrt.equations[eid] for eid in block.eids)
            #
            has_no_qids = (not block_level_qids) and (not block_change_qids)
            has_no_equations = not block_equations
            #
            solve = getattr(_solver_dispatcher, solver)
            if has_no_qids or has_no_equations:
                info_b["success"] = True
                info_b["exit_status"] = _ne.ExitStatus.NO_SOLVER_NEEDED
                info_v["blocks"].append(info_b, )
                print(f"\n-Skipping {custom_header}")
                continue
            #
            steady_evaluator = evaluator_class(
                block_level_qids,
                block_change_qids,
                block_equations,
                all_quantities,
                variant,
                context=self._invariant._context,
                iter_printer_settings=iter_printer_settings | {"custom_header": custom_header},
            )
            #
            maybelog_init_guess = steady_evaluator.get_init_guess()
            maybelog_final_guess, success, exit_status = solve(
                steady_evaluator,
                maybelog_init_guess,
                solver_settings=solver_settings,
            )
            steady_evaluator.final_guess = maybelog_final_guess
            #
            if not success:
                _throw_block_error(human_block, custom_header, )
            #
            # Update variant with steady levels and changes
            _update_variant_with_final_guess(variant, steady_evaluator, qid_to_kind, qid_to_name, )
            #
            info_b["success"] = success
            info_b["exit_status"] = exit_status
            info_v["blocks"].append(info_b, )
        #=======================================================================
        #
        info_v["success"] = all(i["success"] for i in info_v["blocks"])
        return info_v

    _steady_nonlinear_flat = _ft.partialmethod(
        _steady_nonlinear,
        evaluator_class=_evaluators.FlatSteadyEvaluator,
    )

    _steady_nonlinear_nonflat = _ft.partialmethod(
        _steady_nonlinear,
        evaluator_class=_evaluators.NonflatSteadyEvaluator,
    )

    def _resolve_steady_wrt(
        self,
        plan: SteadyPlan,
        is_flat: bool,
    ) -> _Wrt:
        """
        """
        wrt_equations = self.get_steady_equations(kind=_STEADY_EQUATION_SOLVED, )
        wrt_eids = tuple(e.id for e in wrt_equations)
        #
        plannable = self.get_steady_plannable(is_flat=is_flat, )
        wrt_names = set(plannable.can_be_exogenized)
        if plan is None or plan.is_empty:
            exogenized_names = ()
            endogenized_names = ()
            fixed_level_names = ()
            fixed_change_names = ()
        else:
            exogenized_names = plan.get_exogenized_names()
            endogenized_names = plan.get_endogenized_names()
            fixed_level_names = plan.get_fixed_level_names()
            fixed_change_names = plan.get_fixed_change_names() + endogenized_names
        #
        wrt_names = (wrt_names - set(exogenized_names)) | set(endogenized_names)
        # wrt_level_names = (wrt_names - set(fixed_level_names))
        # wrt_change_names = (wrt_names - set(fixed_change_names)) if not is_flat else ()
        # wrt_any_names = wrt_level_names | wrt_change_names
        #
        name_to_qid = self.create_name_to_qid()
        # wrt_level_qids = tuple(sorted([name_to_qid[name] for name in wrt_level_names]))
        # wrt_change_qids = tuple(sorted([name_to_qid[name] for name in wrt_change_names]))
        def get_sorted_qids(names: Iterable[str], ) -> tuple[int, ...]:
            return tuple(sorted([name_to_qid[name] for name in names]))
        wrt_qids = get_sorted_qids(wrt_names, )
        wrt_fixed_level_qids = get_sorted_qids(fixed_level_names, )
        wrt_fixed_change_qids = get_sorted_qids(fixed_change_names, )
        #
        # Make order of names consistent with qids
        qid_to_name = self.create_qid_to_name()
        # wrt_level_names = tuple(qid_to_name[qid] for qid in wrt_level_qids)
        # wrt_change_names = tuple(qid_to_name[qid] for qid in wrt_change_qids)
        wrt_names = tuple(qid_to_name[qid] for qid in wrt_qids)
        wrt_fixed_level_names = tuple(qid_to_name[qid] for qid in wrt_fixed_level_qids)
        wrt_fixed_change_names = tuple(qid_to_name[qid] for qid in wrt_fixed_change_qids)
        #
        return _Wrt(
            equations=wrt_equations,
            eids=wrt_eids,
            names=wrt_names,
            fixed_level_names=wrt_fixed_level_names,
            fixed_change_names=wrt_fixed_change_names,
            qids=wrt_qids,
            fixed_level_qids=wrt_fixed_level_qids,
            fixed_change_qids=wrt_fixed_change_qids,
        )

    def _choose_steady_solver(
        self,
        is_linear: bool,
        is_flat: bool,
        /,
    ) -> Callable:
        """
        Choose steady solver depending on linear and flat flags
        """
        match (is_linear, is_flat, ):
            case (False, False, ):
                return self._steady_nonlinear_nonflat
            case (False, True, ):
                return self._steady_nonlinear_flat
            case (True, False, ):
                return self._steady_linear_nonflat
            case (True, True, ):
                return self._steady_linear_flat

    def check_steady(
        self,
        /,
        equation_switch: Literal["steady", "dynamic"] = "dynamic",
        when_fails: _wrongdoings.HOW = "error",
        tolerance: float = 1e-12,
        return_info: bool = False,
        unpack_singleton: bool = True,
    ) -> tuple[bool, tuple[dict, ...]]:
        """
        Verify currently assigned steady state in dynamic or steady equations for each variant within this model
        """
        qid_to_logly = self.create_qid_to_logly()
        equations = getattr(self._invariant, f"{equation_switch}_equations", )
        equator = self._choose_plain_equator(equation_switch, )
        steady_arrays = (
            v.create_steady_array(
                qid_to_logly,
                num_columns=equator.min_num_columns + 1,
                shift_in_first_column=equator.min_shift,
            ) for v in self._variants
        )
        #
        # REFACTOR
        #
        t_zero = -equator.min_shift
        discrepancies = []
        for x in steady_arrays:
            at_t_zero = equator.eval(x, t_zero, )
            at_t_zero_plus_one = equator.eval(x, t_zero + 1, )
            discrepancies.append(_np.hstack((
                _np.array(at_t_zero).reshape(-1, 1),
                _np.array(at_t_zero_plus_one).reshape(-1, 1),
            )))
        #
        # REFACTOR
        #
        fail_stream = _wrongdoings.create_stream(
            when_fails,
            "Steady state discrepancies in these equations",
        )
        status = []
        for vid, i in enumerate(discrepancies, ):
            where = _np.any(_np.abs(i) > tolerance, axis=1, ).nonzero()[0].tolist()
            for j in where:
                fail_stream.add(f"[Variant {vid}] {equations[j].human}")
            status.append(not where)
        fail_stream._raise()
        all_status = all(status)
        #
        if not return_info:
            return all_status
        #
        info = [
            {"discrepancies": d, }
            for d in discrepancies
        ]
        info = _has_variants.unpack_singleton(
            info, self.is_singleton,
            unpack_singleton=unpack_singleton,
        )
        return all_status, info,

    #]


def _apply_delog_on_vector(
    vector: _np.ndarray,
    qids: Iterable[int],
    qid_to_logly: dict[int, bool],
    /,
) -> None:
    """
    Delogarithmize the elements of numpy vector that have True log-status
    """
    #[
    #]


def _calculate_steady_incidence_matrix(
    equations: Iterable[Equation],
    wrt_any_qids: Iterable[int],
    /,
) -> _np.ndarray:
    """
    """
    qid_to_column = { qid: column for column, qid in enumerate(wrt_any_qids) }
    def token_within_quantities(tok: Token, /, ) -> bool:
        return qid_to_column.get(tok.qid, None)
        #
    return _equations.calculate_incidence_matrix(
        equations,
        len(wrt_any_qids),
        token_within_quantities,
    )


def _resolve_split_into_blocks(
    split_into_blocks: bool | None,
    plan: SteadyPlan | None,
    /,
) -> bool:
    """
    """
    #[
    if split_into_blocks is not None:
        return split_into_blocks
    if plan is None:
        return True
    any_fix = (
        plan.any_in_register("fixed_level", )
        or plan.any_in_register("fixed_change", )
    )
    return not any_fix
    #]


def _update_variant_with_final_guess(variant, steady_evaluator, qid_to_kind, qid_to_name, ) -> None:
    """
    """
    #[
    final_guess = steady_evaluator.final_guess
    #
    levels, wrt_level_qids = steady_evaluator.extract_levels(final_guess, )
    variant.update_levels_from_array(levels, wrt_level_qids, )
    #
    changes, wrt_change_qids = steady_evaluator.extract_changes(final_guess, )
    #
    # Filter out non-loggable quantities, i.e. endogenized parameters (only loggable
    # quantities can have steady-state change)
    changes_to_update = []
    wrt_change_qids_to_update = []
    for change, qid in zip(changes, wrt_change_qids, ):
        if qid_to_kind[qid] not in QuantityKind.LOGGABLE_VARIABLE:
            continue
        changes_to_update.append(change)
        wrt_change_qids_to_update.append(qid)
    variant.update_changes_from_array(changes_to_update, wrt_change_qids_to_update, )
    #]


def _throw_block_error(human_block, custom_header: str, ) -> NoReturn:
    """
    """
    #[
    message = (
        f"Steady state calculations failed to converge in {custom_header}",
        *human_block.equations,
        "Solved for " + " ".join(human_block.quantities, ),
    )
    raise _wrongdoings.IrisPieError(message, )
    #]





"""
First-order stacked-time system
"""


#[

from __future__ import annotations

from .. import has_variants as _has_variants
from ..simultaneous.main import Simultaneous
from .. import quantities as _quantities
from ..quantities import QuantityKind, Quantity
from ..databoxes.main import Databox
from ..dataslates.main import Dataslate
from .. import dates as _dates
from ..dates import Period

from ._invariants import Invariant
from ._variants import Variant
from ._slatable_protocols import Slatable

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..simultaneous.main import Simultaneous

#]


__all__ = [
    "Stacker",
]


class Stacker(
    _has_variants.Mixin,
    _quantities.Mixin,
):
    """
    """
    #[

    __slots__ = (
        "_invariant",
        "_variants",
    )

    @classmethod
    def from_simultaneous(
        klass,
        model: Simultaneous,
        num_periods: int,
        **kwargs,
    ) -> Self:
        """
        """
        self = klass()
        self._invariant = Invariant.from_simultaneous(model, num_periods, **kwargs, )
        self._variants = [
            Variant.from_solution(self._invariant, solution, )
            for solution in model.get_solution(unpack_singleton=False, )
        ]
        return self

    def _access_quantities(self, /, ) -> list[Quantity]:
        """
        Implement quantities.AccessQuantitiesProtocol
        """
        return self._invariant.quantities

    @property
    def max_lag(self, /, ) -> int:
        return self._invariant.max_lag

    @property
    def max_lead(self, /, ) -> int:
        return self._invariant.max_lead

    def get_base_periods(self, start: Period, /, ) -> list[Period]:
        return self._invariant.get_base_periods(start, )

    def create_dataslate(
        self,
        databox: Databox,
        start: Period,
        num_variants: int | None = None,
        shocks_from_data: bool = False,
        # stds_from_data: bool = False,
        # parameters_from_data: bool = False,
    ) -> Dataslate:
        """
        """
        num_variants = self.resolve_num_variants_in_context(num_variants, )
        base_periods = self.get_base_periods(start, )
        #
        slatable = Slatable(
            self,
            shocks_from_data=shocks_from_data,
        )
        #
        dataslate = Dataslate.from_databox_for_slatable(
            slatable, databox, base_periods,
            num_variants=num_variants,
        )
        #
        return dataslate

    #]

